"""Validation of configuration."""
import logging
import os
from pathlib import Path
from schema import Or, Schema, SchemaError

from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)

# pylint: disable=line-too-long


def validate_config(config: dict):
    """Validate configuration.

    Parameters
    ----------
    config: dict
        Config dictionary imported by read_yaml() and parsed through clean_config().
    """
    config_schema = Schema(
        {
            "base_dir": Path,
            "output_dir": Path,
            "warnings": Or("ignore", error="Invalid value in config for 'warnings', valid values are 'ignore'"),
            "cores": lambda n: 1 <= n <= os.cpu_count(),
            "quiet": Or(True, False, error="Invalid value in config for 'quiet', valid values are 'True' or 'False'"),
            "file_ext": Or(
                ".spm",
                ".jpk",
                ".ibw",
                error="Invalid value in config for 'file_ext', valid values are '.spm', '.jpk' or '.ibw'",
            ),
            "loading": {
                "channel": Or(
                    "ZSensor",
                    "",
                    "Stiffness",
                    "LogStiffness",
                    "Adhesion",
                    "Deformation",
                    "Dissipation",
                    "Height",  # end of spm channels
                    "HeightTracee",
                    "HeightRetrace",
                    "ZSensorTrace",
                    "ZSensorRetrace",
                    "UserIn0Trace",
                    "UserIn0Retrace",
                    "UserIn1Trace",
                    "UserIn1Retrace",  # end of ibw channels
                    error="Invalid value in config file for 'channel', all possible"
                    "image channels are seen in the above error message.",
                )
            },
            "filter": {
                "run": Or(
                    True,
                    False,
                    error="Invalid value in config for 'filter.run', valid values are 'True' or 'False'",
                ),
                "threshold_method": Or(
                    "absolute",
                    "otsu",
                    "std_dev",
                    error=(
                        "Invalid value in config for 'filter.threshold_method', valid values "
                        "are 'absolute', 'otsu' or 'std_dev'"
                    ),
                ),
                "otsu_threshold_multiplier": float,
                "threshold_std_dev": {
                    "lower": lambda n: n > 0,
                    "upper": lambda n: n > 0,
                },
                "threshold_absolute": {
                    "lower": lambda n: n < 0,
                    "upper": lambda n: n > 0,
                },
                "gaussian_size": float,
                "gaussian_mode": Or(
                    "nearest",
                    error="Invalid value in config for 'filter.gaussian_mode', valid values are 'nearest'",
                ),
            },
            "grains": {
                "run": Or(
                    True, False, error="Invalid value in config for grains.run, valid values are 'True' or 'False'"
                ),
                "absolute_smallest_grain_size": int,
                "threshold_method": Or(
                    "absolute",
                    "otsu",
                    "std_dev",
                    error=(
                        "Invalid value in config for 'grains.threshold_method', valid values "
                        "are 'absolute', 'otsu' or 'std_dev'"
                    ),
                ),
                "otsu_threshold_multiplier": float,
                "threshold_std_dev": {
                    "lower": lambda n: n > 0,
                    "upper": lambda n: n > 0,
                },
                "threshold_absolute": {
                    "lower": lambda n: n < 0,
                    "upper": lambda n: n > 0,
                },
                "absolute_area_threshold": {
                    "upper": [
                        Or(
                            int,
                            None,
                            error=(
                                "Invalid value in config for 'grains.absolute_area_threshold.upper', valid values "
                                "are int or null"
                            ),
                        )
                    ],
                    "lower": [
                        Or(
                            int,
                            None,
                            error=(
                                "Invalid value in config for 'grains.absolute_area_threshold.lower', valid values "
                                "are int or null"
                            ),
                        )
                    ],
                },
                "direction": Or(
                    "both",
                    "lower",
                    "upper",
                    error="Invalid direction for grains.direction valid values are 'both', 'lower' or 'upper",
                ),
            },
            "grainstats": {
                "run": Or(
                    True,
                    False,
                    error="Invalid value in config for 'grainstats.run', valid values are 'True' or 'False'",
                ),
                "cropped_size": Or(
                    float,
                    int,
                ),
                "save_cropped_grains": Or(
                    True,
                    False,
                    error=(
                        "Invalid value in config for 'grainstats.save_cropped_grains, valid values "
                        "are 'True' or 'False'"
                    ),
                ),
            },
            "dnatracing": {
                "run": Or(
                    True,
                    False,
                    error="Invalid value in config for 'filter.run', valid values are 'True' or 'False'",
                )
            },
            "plotting": {
                "run": Or(
                    True,
                    False,
                    error="Invalid value in config for 'plotting.run', valid values are 'True' or 'False'",
                ),
                "save_format": str,
                "image_set": Or(
                    "all",
                    "core",
                    error="Invalid value in config for 'plotting.image_set', valid values " "are 'all' or 'core'",
                ),
                "zrange": list,
                "colorbar": Or(
                    True,
                    False,
                    error="Invalid value in config for 'plotting.colorbar', valid values are 'True' or 'False'",
                ),
                "axes": Or(
                    True,
                    False,
                    error="Invalid value in config plotting.for 'axes', valid values are 'True' or 'False'",
                ),
                "cmap": Or(
                    "afmhot",
                    "nanoscope",
                    error="Invalid value in config for 'plotting.cmap', valid values are 'afmhot' or 'nanoscope",
                ),
                "histogram_log_axis": Or(
                    True,
                    False,
                    error="Invalid value in config plotting histogram. For 'log_y_axis', valid values are 'True' or 'False'",
                ),
            },
        }
    )

    try:
        config_schema.validate(config)
        LOGGER.info("Configuration is valid.")
    except SchemaError as schema_error:
        raise SchemaError(
            "There is an error in your configuration. Please refer to the first error message above for details"
        ) from schema_error


def validate_plotting(config: dict) -> None:
    """Validate configuration.

    Parameters
    ----------
    config: dict
        Config dictionary imported by read_yaml() and parsed through clean_config().
    """
    config_schema = Schema(
        {
            "extracted_channel": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'extracted_channel.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "pixels": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error="Invalid value in config 'pixels.image_type', valid values are 'binary' or 'non-binary'",
                ),
                "core_set": bool,
            },
            "initial_median_flatten": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'initial_median_flatten.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "initial_tilt_removal": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'initial_tilt_removal.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "initial_quadratic_removal": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'initial_quadratic_removal.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "mask": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error="Invalid value in config 'mask.image_type', valid values are 'binary' or 'non-binary'",
                ),
                "core_set": bool,
            },
            "masked_median_flatten": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'masked_median_flatten.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "masked_tilt_removal": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'masked_tilt_removal.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "masked_quadratic_removal": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'masked_quadratic_removal.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "zero_average_background": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'gaussian_filtered.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "gaussian_filtered": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'gaussian_filtered.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "z_threshed": {
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'z_threshold.image_type', valid values " "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": True,
            },
            "mask_grains": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'mask_grains.image_type', valid values " "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "labelled_regions_01": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'labelled_regions_01.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "tidied_border": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'tidied_border.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "removed_noise": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'removed_noise.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "removed_small_objects": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'removed_small_objects.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "mask_overlay": {
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'mask_overlay.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": True,
            },
            "labelled_regions_02": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'labelled_regions_02.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "coloured_regions": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'coloured_regions.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "bounding_boxes": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'bounding_boxes.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "coloured_boxes": {
                "filename": str,
                "title": str,
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'coloured_boxes.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "grain_image": {
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'grain_image.image_type', valid values " "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": True,
            },
            "grain_mask": {
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'grain_mask.image_type', valid values " "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
            "grain_mask_image": {
                "image_type": Or(
                    "binary",
                    "non-binary",
                    error=(
                        "Invalid value in config 'grain_mask_image.image_type', valid values "
                        "are 'binary' or 'non-binary'"
                    ),
                ),
                "core_set": bool,
            },
        }
    )
    try:
        config_schema.validate(config)
        LOGGER.info("Plotting configuration is valid.")
    except SchemaError as schema_error:
        raise SchemaError(
            "There is an error in your configuration. Please refer to the first error message above for details"
        ) from schema_error



"""Plotting data."""
from pathlib import Path
from typing import Union
import logging

from matplotlib.patches import Rectangle, Patch
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np

from topostats.logs.logs import LOGGER_NAME
from topostats.theme import Colormap

LOGGER = logging.getLogger(LOGGER_NAME)


class Images:
    """Plots image arrays"""

    def __init__(
        self,
        data: np.array,
        output_dir: Union[str, Path],
        filename: str,
        pixel_to_nm_scaling: float = 1.0,
        data2: np.array = None,
        title: str = None,
        image_type: str = "non-binary",
        image_set: str = "core",
        core_set: bool = False,
        interpolation: str = "nearest",
        cmap: str = "nanoscope",
        region_properties: dict = None,
        zrange: list = [None, None],
        colorbar: bool = True,
        axes: bool = True,
        save: bool = True,
        save_format: str = "png",
        histogram_log_axis: bool = True,
    ) -> None:
        """
        Initialise the class.

        Parameters
        ----------
        data : np.array
            Numpy array to plot.
        output_dir : Union[str, Path]
            Output directory to save the file to.
        filename : Union[str, Path]
            Filename to save image as.
        pixel_to_nm_scaling : float
            The scaling factor showing the real length of 1 pixel, in nm.
        data2 : np.ndarray
            Optional mask array to overlay onto an image.
        title : str
            Title for plot.
        image_type : str
            The image data type - binary or non-binary.
        image_set : str
            The set of images to process - core or all.
        core_set : bool
            Flag to identify image as part of the core image set or not.
        interpolation: str
            Interpolation to use (default 'nearest').
        cmap : str
            Colour map to use (default 'nanoscope', 'afmhot' also available)
        region_properties: dict
            Dictionary of region properties, adds bounding boxes if specified.
        zrange : list
            Lower and upper bound to clip core images to.
        colorbar: bool
            Optionally add a colorbar to plots, default is False.
        axes: bool
            Optionally add/remove axes from the image.
        save: bool
            Whether to save the image.
        save_format: str
            Format to save the image as.
        histogram_log_axis: bool
            Optionally use a logarithmic y axis for the histogram plots
        """

        self.data = data
        self.output_dir = Path(output_dir)
        self.filename = filename
        self.pixel_to_nm_scaling = pixel_to_nm_scaling
        self.data2 = data2
        self.title = title
        self.image_type = image_type
        self.image_set = image_set
        self.core_set = core_set
        self.interpolation = interpolation
        self.cmap = cmap
        self.region_properties = region_properties
        self.zrange = zrange
        self.colorbar = colorbar
        self.axes = axes
        self.save = save
        self.save_format = save_format
        self.histogram_log_axis = histogram_log_axis

    def plot_histogram_and_save(self):
        """
        Plot and save a histogram of the height map

        Returns
        -------
        fig: plt.figure.Figure
            Matplotlib.pyplot figure object
        ax: plt.axes._subplots.AxesSubplot
            Matplotlib.pyplot axes object
        """
        if self.image_set == "all":
            fig, ax = plt.subplots(1, 1, figsize=(8, 8))

            ax.hist(self.data.flatten().astype(float), bins="auto", log=self.histogram_log_axis)
            ax.set_xlabel("pixel height")
            if self.histogram_log_axis:
                ax.set_ylabel("frequency in image (log)")
            else:
                ax.set_ylabel("frequency in image")
            plt.title(self.title)
            plt.savefig(
                (self.output_dir / f"{self.filename}_histogram.{self.save_format}"),
                format=self.save_format,
                bbox_inches="tight",
                pad_inches=0.5,
            )
            plt.close()

            return fig, ax
        else:
            return None

    def plot_and_save(self):
        """
        Plot and save the images with savefig or imsave depending on config file parameters.

        Returns
        -------
        fig: plt.figure.Figure
            Matplotlib.pyplot figure object
        ax: plt.axes._subplots.AxesSubplot
            Matplotlib.pyplot axes object
        """
        fig, ax = None, None
        if self.save:
            if self.image_set == "all" or self.core_set:
                if self.axes or self.colorbar:
                    fig, ax = self.save_figure()
                else:
                    if isinstance(self.data2, np.ndarray) or self.region_properties:
                        fig, ax = self.save_figure()
                    else:
                        self.save_array_figure()
        if "_processed" in self.filename:
            LOGGER.info(
                f"[{self.filename.split('_processed')[0]}] : Image saved to : {str(self.output_dir / self.filename)}"
            )
        return fig, ax

    def save_figure(self):
        """
        This function saves figures as plt.savefig objects.

        Returns
        -------
        fig: plt.figure.Figure
            Matplotlib.pyplot figure object
        ax: plt.axes._subplots.AxesSubplot
            Matplotlib.pyplot axes object
        """
        fig, ax = plt.subplots(1, 1, figsize=(8, 8))
        shape = self.data.shape
        if isinstance(self.data, np.ndarray):
            im = ax.imshow(
                self.data,
                extent=(0, shape[1] * self.pixel_to_nm_scaling, 0, shape[0] * self.pixel_to_nm_scaling),
                interpolation=self.interpolation,
                cmap=Colormap(self.cmap).get_cmap(),
                vmin=self.zrange[0],
                vmax=self.zrange[1],
            )
            if isinstance(self.data2, np.ndarray):
                self.data2[self.data2 != 0] = 1
                mask = np.ma.masked_where(self.data2 == 0, self.data2)
                ax.imshow(
                    mask,
                    "jet_r",
                    extent=(
                        0,
                        shape[1] * self.pixel_to_nm_scaling,
                        0,
                        shape[0] * self.pixel_to_nm_scaling,
                    ),
                    interpolation=self.interpolation,
                    alpha=0.7,
                )
                patch = [Patch(color=plt.get_cmap("jet_r")(1, 0.7), label="Mask")]
                plt.legend(handles=patch, loc="upper right", bbox_to_anchor=(1, 1.06))

            plt.title(self.title)
            plt.xlabel("Nanometres")
            plt.ylabel("Nanometres")
            plt.axis(self.axes)
            if self.colorbar and self.image_type == "non-binary":
                divider = make_axes_locatable(ax)
                cax = divider.append_axes("right", size="5%", pad=0.05)
                plt.colorbar(im, cax=cax, label="Height (Nanometres)")
            if self.region_properties:
                fig, ax = add_bounding_boxes_to_plot(fig, ax, shape, self.region_properties, self.pixel_to_nm_scaling)
            if not self.axes and not self.colorbar:
                plt.title("")
                fig.frameon = False
                plt.savefig(
                    (self.output_dir / f"{self.filename}.{self.save_format}"),
                    format=self.save_format,
                    bbox_inches="tight",
                    pad_inches=0,
                )
            else:
                plt.savefig((self.output_dir / f"{self.filename}.{self.save_format}"), format=self.save_format)
        else:
            plt.xlabel("Nanometres")
            plt.ylabel("Nanometres")
            self.data.show(
                ax=ax,
                extent=(0, shape[1] * self.pixel_to_nm_scaling, 0, shape[0] * self.pixel_to_nm_scaling),
                interpolation=self.interpolation,
                cmap=Colormap(self.cmap).get_cmap(),
            )
        plt.close()
        return fig, ax

    def save_array_figure(self) -> None:
        """This function saves only the image array as an image using plt.imsave"""
        plt.imsave(
            (self.output_dir / f"{self.filename}.{self.save_format}"),
            self.data,
            cmap=Colormap(self.cmap).get_cmap(),
            vmin=self.zrange[0],
            vmax=self.zrange[1],
            format=self.save_format,
        )
        plt.close()


def add_bounding_boxes_to_plot(fig, ax, shape, region_properties: list, pixel_to_nm_scaling: float) -> None:
    """Add the bounding boxes to a plot.

    Parameters
    ----------
    fig: plt.figure.Figure
        Matplotlib.pyplot figure object
    ax: plt.axes._subplots.AxesSubplot.
        Matplotlib.pyplot axes object
    shape: tuple
        Tuple of the image-to-be-plot's shape.
    region_properties:
        Region properties to add bounding boxes from.
    pixel_to_nm_scaling: float
        The scaling factor from px to nm.

    Returns
    -------
    fig: plt.figure.Figure
        Matplotlib.pyplot figure object.
    ax: plt.axes._subplots.AxesSubplot
        Matplotlib.pyplot axes object.
    """
    for region in region_properties:
        min_y, min_x, max_y, max_x = [x * pixel_to_nm_scaling for x in region.bbox]
        # Correct y-axis
        min_y = (shape[0] * pixel_to_nm_scaling) - min_y
        max_y = (shape[0] * pixel_to_nm_scaling) - max_y
        rectangle = Rectangle((min_x, min_y), max_x - min_x, max_y - min_y, fill=False, edgecolor="white", linewidth=2)
        ax.add_patch(rectangle)
    return fig, ax



"""Perform DNA Tracing"""
from collections import OrderedDict
import logging
from pathlib import Path
import math
import os
from typing import Union, Tuple
import warnings

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from scipy import ndimage, spatial, interpolate as interp
from skimage import morphology
from skimage.filters import gaussian

from topostats.logs.logs import LOGGER_NAME
from topostats.tracing.tracingfuncs import genTracingFuncs, getSkeleton, reorderTrace

LOGGER = logging.getLogger(LOGGER_NAME)


class dnaTrace(object):
    """
    This class gets all the useful functions from the old tracing code and staples
    them together to create an object that contains the traces for each DNA molecule
    in an image and functions to calculate stats from those traces.

    The traces are stored in dictionaries labelled by their gwyddion defined grain
    number and are represented as numpy arrays.

    The object also keeps track of the skeletonised plots and other intermediates
    in case these are useful for other things in the future.
    """

    def __init__(
        self,
        full_image_data,
        grains,
        filename,
        pixel_size,
        convert_nm_to_m: bool = True,
    ):
        self.full_image_data = full_image_data * 1e-9 if convert_nm_to_m else full_image_data
        # self.grains_orig = [x for row in grains for x in row]
        self.grains_orig = grains
        self.filename = filename
        self.pixel_size = pixel_size * 1e-9 if convert_nm_to_m else pixel_size
        # self.number_of_columns = number_of_columns
        # self.number_of_rows = number_of_rows
        self.number_of_rows = self.full_image_data.shape[0]
        self.number_of_columns = self.full_image_data.shape[1]
        self.sigma = 0.7 / (self.pixel_size * 1e9)

        self.gauss_image = []
        self.grains = {}
        self.dna_masks = {}
        self.skeletons = {}
        self.disordered_trace = {}
        self.ordered_traces = {}
        self.fitted_traces = {}
        self.splined_traces = {}
        self.contour_lengths = {}
        self.end_to_end_distance = {}
        self.mol_is_circular = {}
        self.curvature = {}

        self.number_of_traces = 0
        self.num_circular = 0
        self.num_linear = 0

        self.neighbours = 5  # The number of neighbours used for the curvature measurement

        # supresses scipy splining warnings
        warnings.filterwarnings("ignore")

        LOGGER.debug(f"[{self.filename}] Performing DNA Tracing")

    def trace_dna(self):
        """Perform DNA tracing."""
        self.get_numpy_arrays()
        self.gaussian_filter()
        self.get_disordered_trace()
        # self.isMolLooped()
        self.purge_obvious_crap()
        self.linear_or_circular(self.disordered_trace)
        self.get_ordered_traces()
        self.linear_or_circular(self.ordered_traces)
        self.get_fitted_traces()
        self.get_splined_traces()
        # self.find_curvature()
        # self.saveCurvature()
        self.measure_contour_length()
        self.measure_end_to_end_distance()
        self.report_basic_stats()

    def get_numpy_arrays(self):

        """Function to get each grain as a numpy array which is stored in a
        dictionary

        Currently the grains are unnecessarily large (the full image) as I don't
        know how to handle the cropped versions

        I find using the gwyddion objects clunky and not helpful once the
        grains have been found

        There is some kind of discrepency between the ordering of arrays from
        gwyddion and how they're usually handled in np arrays meaning you need
        to be careful when indexing from gwyddion derived numpy arrays"""
        for grain_num in set(self.grains_orig.flatten()):
            # Skip the background
            if grain_num != 0:
                # Saves each grain as a multidim numpy array
                # single_grain_1d = np.array([1 if i == grain_num else 0 for i in self.grains_orig])
                # self.grains[int(grain_num)] = np.reshape(single_grain_1d, (self.number_of_columns, self.number_of_rows))
                self.grains[int(grain_num)] = self._get_grain_array(grain_num)
        # FIXME : This should be a method of its own, but strange that apparently Gaussian filtered image is filtered again
        # Get a 7 A gauss filtered version of the original image
        # used in refining the pixel positions in fitted_traces()
        # sigma = 0.7 / (self.pixel_size * 1e9)
        # self.gauss_image = filters.gaussian(self.full_image_data, sigma)

    def _get_grain_array(self, grain_number: int) -> np.ndarray:
        """Extract a single grains."""
        return np.where(self.grains_orig == grain_number, 1, 0)

    # FIXME : It is straight-forward to get bounding boxes for grains, need to then have a dictionary of original image
    #         and label for each grain to then be processed.
    def _get_bounding_box(array: np.ndarray) -> np.ndarray:
        """Calculate bounding box for each grain."""
        rows = grain_array.any(axis=1)
        LOGGER.info(f"[{self.filename}] : Cropping grain")
        if rows.any():
            m, n = grain_array.shape
            cols = grain_array.any(0)
            return (rows.argmax(), m - rows[::-1].argmax(), cols.argmax(), n - cols[::-1].argmax())
            return grain_array[rows.argmax() : m - rows[::-1].argmax(), cols.argmax() : n - cols[::-1].argmax()]
        return np.empty((0, 0), dtype=bool)

    def _crop_array(array: np.ndarray, bounding_box: Tuple) -> np.ndarray:
        """Crop an array.

        Parameters
        ----------
        array: np.ndarray
            2D Numpy array to be cropped.
        bounding_box: Tuple
            Tuple of co-ordinates to crop, should be of form (max_x, min_x, max_y, min_y) as returned by
        _get_bounding_box().

        Returns
        -------
        np.ndarray()
            Cropped array
        """
        return array[bounding_box[0] : bounding_box[1], bounding_box[2] : bounding_box[3]]

    def gaussian_filter(self, **kwargs) -> np.array:
        """Apply Gaussian filter"""
        self.gauss_image = gaussian(self.full_image_data, sigma=self.sigma, **kwargs)
        LOGGER.info(f"[{self.filename}] : Gaussian filter applied.")

    def get_disordered_trace(self):
        """Create a skeleton for each of the grains in the image.

        Uses my own skeletonisation function from tracingfuncs module. I will
        eventually get round to editing this function to try to reduce the branching
        and to try to better trace from looped molecules"""

        for grain_num in sorted(self.grains.keys()):
            # What purpose does binary dilation serve here? Name suggests some sort of smoothing around the edges and
            # the resulting array is used as a mask during the skeletonising process.
            smoothed_grain = ndimage.binary_dilation(self.grains[grain_num], iterations=1).astype(
                self.grains[grain_num].dtype
            )

            sigma = 0.01 / (self.pixel_size * 1e9)
            very_smoothed_grain = ndimage.gaussian_filter(smoothed_grain, sigma)

            try:
                dna_skeleton = getSkeleton(
                    self.gauss_image, smoothed_grain, self.number_of_columns, self.number_of_rows, self.pixel_size
                )
                self.disordered_trace[grain_num] = dna_skeleton.output_skeleton
            except IndexError:
                # Some gwyddion grains touch image border causing IndexError
                # These grains are deleted
                self.grains.pop(grain_num)
            # skel = morphology.skeletonize(self.grains[grain_num])
            # self.skeletons[grain_num] = np.argwhere(skel == 1)

    def purge_obvious_crap(self):

        for dna_num in sorted(self.disordered_trace.keys()):

            if len(self.disordered_trace[dna_num]) < 10:
                self.disordered_trace.pop(dna_num, None)

    def linear_or_circular(self, traces):

        """Determines whether each molecule is circular or linear based on the local environment of each pixel from the trace

        This function is sensitive to branches from the skeleton so might need to implement a function to remove them"""

        self.num_circular = 0
        self.num_linear = 0

        for dna_num in sorted(traces.keys()):

            points_with_one_neighbour = 0
            fitted_trace_list = traces[dna_num].tolist()

            # For loop determines how many neighbours a point has - if only one it is an end
            for x, y in fitted_trace_list:

                if genTracingFuncs.countNeighbours(x, y, fitted_trace_list) == 1:
                    points_with_one_neighbour += 1
                else:
                    pass

            if points_with_one_neighbour == 0:
                self.mol_is_circular[dna_num] = True
                self.num_circular += 1
            else:
                self.mol_is_circular[dna_num] = False
                self.num_linear += 1

    def get_ordered_traces(self):

        for dna_num in sorted(self.disordered_trace.keys()):

            circle_tracing = True

            if self.mol_is_circular[dna_num]:

                self.ordered_traces[dna_num], trace_completed = reorderTrace.circularTrace(
                    self.disordered_trace[dna_num]
                )

                if not trace_completed:
                    self.mol_is_circular[dna_num] = False
                    try:
                        self.ordered_traces[dna_num] = reorderTrace.linearTrace(self.ordered_traces[dna_num].tolist())
                    except UnboundLocalError:
                        self.mol_is_circular.pop(dna_num)
                        self.disordered_trace.pop(dna_num)
                        self.grains.pop(dna_num)
                        self.ordered_traces.pop(dna_num)

            elif not self.mol_is_circular[dna_num]:
                self.ordered_traces[dna_num] = reorderTrace.linearTrace(self.disordered_trace[dna_num].tolist())

    def report_basic_stats(self):
        """Report number of circular and linear DNA molecules detected."""
        LOGGER.info(
            f"There are {self.num_circular} circular and {self.num_linear} linear DNA molecules found in the image"
        )

    def get_fitted_traces(self):
        """Create trace coordinates (for each identified molecule) that are adjusted to lie
        along the highest points of each traced molecule
        """

        for dna_num in sorted(self.ordered_traces.keys()):

            individual_skeleton = self.ordered_traces[dna_num]

            # This indexes a 3 nm height profile perpendicular to DNA backbone
            # note that this is a hard coded parameter
            index_width = int(3e-9 / (self.pixel_size))
            if index_width < 2:
                index_width = 2

            for coord_num, trace_coordinate in enumerate(individual_skeleton):
                height_values = None

                # Block of code to prevent indexing outside image limits
                # e.g. indexing self.gauss_image[130, 130] for 128x128 image
                if trace_coordinate[0] < 0:
                    # prevents negative number indexing
                    # i.e. stops (trace_coordinate - index_width) < 0
                    trace_coordinate[0] = index_width
                elif trace_coordinate[0] >= (self.number_of_rows - index_width):
                    # prevents indexing above image range causing IndexError
                    trace_coordinate[0] = self.number_of_rows - index_width
                # do same for y coordinate
                elif trace_coordinate[1] < 0:
                    trace_coordinate[1] = index_width
                elif trace_coordinate[1] >= (self.number_of_columns - index_width):
                    trace_coordinate[1] = self.number_of_columns - index_width

                # calculate vector to n - 2 coordinate in trace
                if self.mol_is_circular[dna_num]:
                    nearest_point = individual_skeleton[coord_num - 2]
                    vector = np.subtract(nearest_point, trace_coordinate)
                    vector_angle = math.degrees(math.atan2(vector[1], vector[0]))
                else:
                    try:
                        nearest_point = individual_skeleton[coord_num + 2]
                    except IndexError:
                        nearest_point = individual_skeleton[coord_num - 2]
                    vector = np.subtract(nearest_point, trace_coordinate)
                    vector_angle = math.degrees(math.atan2(vector[1], vector[0]))

                if vector_angle < 0:
                    vector_angle += 180

                # if  angle is closest to 45 degrees
                if 67.5 > vector_angle >= 22.5:
                    perp_direction = "negative diaganol"
                    # positive diagonal (change in x and y)
                    # Take height values at the inverse of the positive diaganol
                    # (i.e. the negative diaganol)
                    y_coords = np.arange(trace_coordinate[1] - index_width, trace_coordinate[1] + index_width)[::-1]
                    x_coords = np.arange(trace_coordinate[0] - index_width, trace_coordinate[0] + index_width)

                # if angle is closest to 135 degrees
                elif 157.5 >= vector_angle >= 112.5:
                    perp_direction = "positive diaganol"
                    y_coords = np.arange(trace_coordinate[1] - index_width, trace_coordinate[1] + index_width)
                    x_coords = np.arange(trace_coordinate[0] - index_width, trace_coordinate[0] + index_width)

                # if angle is closest to 90 degrees
                if 112.5 > vector_angle >= 67.5:
                    perp_direction = "horizontal"
                    x_coords = np.arange(trace_coordinate[0] - index_width, trace_coordinate[0] + index_width)
                    y_coords = np.full(len(x_coords), trace_coordinate[1])

                elif 22.5 > vector_angle:  # if angle is closest to 0 degrees
                    perp_direction = "vertical"
                    y_coords = np.arange(trace_coordinate[1] - index_width, trace_coordinate[1] + index_width)
                    x_coords = np.full(len(y_coords), trace_coordinate[0])

                elif vector_angle >= 157.5:  # if angle is closest to 180 degrees
                    perp_direction = "vertical"
                    y_coords = np.arange(trace_coordinate[1] - index_width, trace_coordinate[1] + index_width)
                    x_coords = np.full(len(y_coords), trace_coordinate[0])

                # Use the perp array to index the guassian filtered image
                perp_array = np.column_stack((x_coords, y_coords))
                height_values = self.gauss_image[perp_array[:, 1], perp_array[:, 0]]

                """
                # Old code that interpolated the height profile for "sub-pixel
                # accuracy" - probably slow and not necessary, can delete

                #Use interpolation to get "sub pixel" accuracy for heighest position
                if perp_direction == 'negative diaganol':
                    int_func = interp.interp1d(perp_array[:,0], np.ndarray.flatten(height_values), kind = 'cubic')
                    interp_heights = int_func(np.arange(perp_array[0,0], perp_array[-1,0], 0.1))

                elif perp_direction == 'positive diaganol':
                    int_func = interp.interp1d(perp_array[:,0], np.ndarray.flatten(height_values), kind = 'cubic')
                    interp_heights = int_func(np.arange(perp_array[0,0], perp_array[-1,0], 0.1))

                elif perp_direction == 'vertical':
                    int_func = interp.interp1d(perp_array[:,1], np.ndarray.flatten(height_values), kind = 'cubic')
                    interp_heights = int_func(np.arange(perp_array[0,1], perp_array[-1,1], 0.1))

                elif perp_direction == 'horizontal':
                    #print(perp_array[:,0])
                    #print(np.ndarray.flatten(height_values))
                    int_func = interp.interp1d(perp_array[:,0], np.ndarray.flatten(height_values), kind = 'cubic')
                    interp_heights = int_func(np.arange(perp_array[0,0], perp_array[-1,0], 0.1))
                else:
                    quit('A fatal error occured in the CorrectHeightPositions function, this was likely caused by miscalculating vector angles')

                #Make "fine" coordinates which have the same number of coordinates as the interpolated height values
                if perp_direction == 'negative diaganol':
                    fine_x_coords = np.arange(perp_array[0,0], perp_array[-1,0], 0.1)
                    fine_y_coords = np.arange(perp_array[-1,1], perp_array[0,1], 0.1)[::-1]
                elif perp_direction == 'positive diaganol':
                    fine_x_coords = np.arange(perp_array[0,0], perp_array[-1,0], 0.1)
                    fine_y_coords = np.arange(perp_array[0,1], perp_array[-1,1], 0.1)
                elif perp_direction == 'vertical':
                    fine_y_coords = np.arange(perp_array[0,1], perp_array[-1,1], 0.1)
                    fine_x_coords = np.full(len(fine_y_coords), trace_coordinate[0], dtype = 'float')
                elif perp_direction == 'horizontal':
                    fine_x_coords = np.arange(perp_array[0,0], perp_array[-1,0], 0.1)
                    fine_y_coords = np.full(len(fine_x_coords), trace_coordinate[1], dtype = 'float')
                """
                # Grab x,y coordinates for highest point
                # fine_coords = np.column_stack((fine_x_coords, fine_y_coords))
                sorted_array = perp_array[np.argsort(height_values)]
                highest_point = sorted_array[-1]

                try:
                    # could use np.append() here
                    fitted_coordinate_array = np.vstack((fitted_coordinate_array, highest_point))
                except UnboundLocalError:
                    fitted_coordinate_array = highest_point

            self.fitted_traces[dna_num] = fitted_coordinate_array
            del fitted_coordinate_array  # cleaned up by python anyway?

    def get_splined_traces(self):
        """Gets a splined version of the fitted trace - useful for finding the radius of gyration etc

        This function actually calculates the average of several splines which is important for getting a good fit on
        the lower res data"""

        step_size = int(7e-9 / (self.pixel_size))  # 3 nm step size
        interp_step = int(1e-10 / self.pixel_size)
        # Lets see if we just got with the pixel_to_nm_scaling
        # step_size = self.pixel_size
        # interp_step = self.pixel_size

        # FIXME : Iterate over self.fitted_traces directly use either self.fitted_traces.values() or self.fitted_trace.items()
        for dna_num in sorted(self.fitted_traces.keys()):

            self.splining_success = True
            nbr = len(self.fitted_traces[dna_num][:, 0])

            # Hard to believe but some traces have less than 4 coordinates in total
            if len(self.fitted_traces[dna_num][:, 1]) < 4:
                self.splined_traces[dna_num] = self.fitted_traces[dna_num]
                continue

            # The degree of spline fit used is 3 so there cannot be less than 3 points in the splined trace
            # LOGGER.info(f"DNA Number      : {dna_num}")
            # LOGGER.info(f"nbr             : {nbr}")
            # LOGGER.info(f"step_size       : {step_size}")
            # LOGGER.info(f"self.pixel_size : {self.pixel_size}")
            while nbr / step_size < 4:
                if step_size <= 1:
                    step_size = 1
                    break
                step_size = -1
            if self.mol_is_circular[dna_num]:

                # if nbr/step_size > 4: #the degree of spline fit is 3 so there cannot be less than 3 points in splined trace

                # ev_array = np.linspace(0, 1, nbr * step_size)
                ev_array = np.linspace(0, 1, int(nbr * step_size))

                for i in range(step_size):
                    x_sampled = np.array(
                        [
                            self.fitted_traces[dna_num][:, 0][j]
                            for j in range(i, len(self.fitted_traces[dna_num][:, 0]), step_size)
                        ]
                    )
                    y_sampled = np.array(
                        [
                            self.fitted_traces[dna_num][:, 1][j]
                            for j in range(i, len(self.fitted_traces[dna_num][:, 1]), step_size)
                        ]
                    )

                    try:
                        tck, u = interp.splprep([x_sampled, y_sampled], s=0, per=2, quiet=1, k=3)
                        out = interp.splev(ev_array, tck)
                        splined_trace = np.column_stack((out[0], out[1]))
                    except ValueError:
                        # Value error occurs when the "trace fitting" really messes up the traces

                        x = np.array(
                            [
                                self.ordered_traces[dna_num][:, 0][j]
                                for j in range(i, len(self.ordered_traces[dna_num][:, 0]), step_size)
                            ]
                        )
                        y = np.array(
                            [
                                self.ordered_traces[dna_num][:, 1][j]
                                for j in range(i, len(self.ordered_traces[dna_num][:, 1]), step_size)
                            ]
                        )

                        try:
                            tck, u = interp.splprep([x, y], s=0, per=2, quiet=1)
                            out = interp.splev(np.linspace(0, 1, nbr * step_size), tck)
                            splined_trace = np.column_stack((out[0], out[1]))
                        except ValueError:  # sometimes even the ordered_traces are too bugged out so just delete these traces
                            self.mol_is_circular.pop(dna_num)
                            self.disordered_trace.pop(dna_num)
                            self.grains.pop(dna_num)
                            self.ordered_traces.pop(dna_num)
                            self.splining_success = False
                            try:
                                del spline_running_total
                            except UnboundLocalError:  # happens if splining fails immediately
                                break
                            break

                    try:
                        spline_running_total = np.add(spline_running_total, splined_trace)
                    except NameError:
                        spline_running_total = np.array(splined_trace)

                if not self.splining_success:
                    continue

                spline_average = np.divide(spline_running_total, [step_size, step_size])
                del spline_running_total
                self.splined_traces[dna_num] = spline_average
                # else:
                #    x = self.fitted_traces[dna_num][:,0]
                #    y = self.fitted_traces[dna_num][:,1]

                #    try:
                #        tck, u = interp.splprep([x, y], s=0, per = 2, quiet = 1, k = 3)
                #        out = interp.splev(np.linspace(0,1,nbr*step_size), tck)
                #        splined_trace = np.column_stack((out[0], out[1]))
                #        self.splined_traces[dna_num] = splined_trace
                #    except ValueError: #if the trace is really messed up just delete it
                #        self.mol_is_circular.pop(dna_num)
                #        self.disordered_trace.pop(dna_num)
                #        self.grains.pop(dna_num)
                #        self.ordered_traces.pop(dna_num)

            else:
                """
                start_x = self.fitted_traces[dna_num][0, 0]
                end_x = self.fitted_traces[dna_num][-1, 0]

                for i in range(step_size):
                    x_sampled = np.array([self.fitted_traces[dna_num][:, 0][j] for j in
                                          range(i, len(self.fitted_traces[dna_num][:, 0]), step_size)])
                    y_sampled = np.array([self.fitted_traces[dna_num][:, 1][j] for j in
                                          range(i, len(self.fitted_traces[dna_num][:, 1]), step_size)])

                    interp_f = interp.interp1d(x_sampled, y_sampled, kind='cubic', assume_sorted=False)

                    x_new = np.linspace(start_x, end_x, interp_step)
                    y_new = interp_f(x_new)

                    print(y_new)

                    # tck = interp.splrep(x_sampled, y_sampled, quiet = 0)
                    # out = interp.splev(np.linspace(start_x,end_x, nbr*step_size), tck)
                    splined_trace = np.column_stack((x_new, y_new))

                    try:
                        np.add(spline_running_total, splined_trace)
                    except NameError:
                        spline_running_total = np.array(splined_trace)

                spline_average = spline_running_total
                self.splined_traces[dna_num] = spline_average
                """

                # can't get splining of linear molecules to work yet
                self.splined_traces[dna_num] = self.fitted_traces[dna_num]

    def show_traces(self):

        plt.pcolormesh(self.gauss_image, vmax=-3e-9, vmin=3e-9)
        plt.colorbar()
        for dna_num in sorted(self.disordered_trace.keys()):
            plt.plot(self.ordered_traces[dna_num][:, 0], self.ordered_traces[dna_num][:, 1], markersize=1)
            plt.plot(self.fitted_traces[dna_num][:, 0], self.fitted_traces[dna_num][:, 1], markersize=1)
            plt.plot(self.splined_traces[dna_num][:, 0], self.splined_traces[dna_num][:, 1], markersize=1)
            # print(len(self.skeletons[dna_num]), len(self.disordered_trace[dna_num]))
            # plt.plot(self.skeletons[dna_num][:,0], self.skeletons[dna_num][:,1], 'o', markersize = 0.8)

        plt.show()
        plt.close()

    def saveTraceFigures(
        self, filename: Union[str, Path], channel_name: str, vmaxval, vminval, output_dir: Union[str, Path] = None
    ):

        # if directory_name:
        #     filename_with_ext = self._checkForSaveDirectory(filename_with_ext, directory_name)

        # save_file = filename_with_ext[:-4]

        # vmaxval = 20e-9
        # vminval = -10e-9

        plt.pcolormesh(self.full_image_data, vmax=vmaxval, vmin=vminval)
        plt.colorbar()
        # plt.savefig("%s_%s_originalImage.png" % (save_file, channel_name))
        plt.savefig(output_dir / filename / f"{channel_name}_original.png")
        plt.close()

        # plt.pcolormesh(self.full_image_data, vmax=vmaxval, vmin=vminval)
        # plt.colorbar()
        # for dna_num in sorted(self.splined_traces.keys()):
        #     # disordered_trace_list = self.ordered_traces[dna_num].tolist()
        #     # less_dense_trace = np.array([disordered_trace_list[i] for i in range(0,len(disordered_trace_list),5)])
        #     plt.plot(self.splined_traces[dna_num][:, 0], self.splined_traces[dna_num][:, 1], color='c', linewidth=1.0)
        #     if self.mol_is_circular[dna_num]:
        #         starting_point = 0
        #     else:
        #         starting_point = self.neighbours
        #     length = len(self.curvature[dna_num])
        #     plt.plot(self.splined_traces[dna_num][starting_point, 0],
        #              self.splined_traces[dna_num][starting_point, 1],
        #              color='#D55E00', markersize=3.0, marker=5)
        #     plt.plot(self.splined_traces[dna_num][starting_point + int(length / 6), 0],
        #              self.splined_traces[dna_num][starting_point + int(length / 6), 1],
        #              color='#E69F00', markersize=3.0, marker=5)
        #     plt.plot(self.splined_traces[dna_num][starting_point + int(length / 6 * 2), 0],
        #              self.splined_traces[dna_num][starting_point + int(length / 6 * 2), 1],
        #              color='#F0E442', markersize=3.0, marker=5)
        #     plt.plot(self.splined_traces[dna_num][starting_point + int(length / 6 * 3), 0],
        #              self.splined_traces[dna_num][starting_point + int(length / 6 * 3), 1],
        #              color='#009E74', markersize=3.0, marker=5)
        #     plt.plot(self.splined_traces[dna_num][starting_point + int(length / 6 * 4), 0],
        #              self.splined_traces[dna_num][starting_point + int(length / 6 * 4), 1],
        #              color='#0071B2', markersize=3.0, marker=5)
        #     plt.plot(self.splined_traces[dna_num][starting_point + int(length / 6 * 5), 0],
        #              self.splined_traces[dna_num][starting_point + int(length / 6 * 5), 1],
        #              color='#CC79A7', markersize=3.0, marker=5)
        # plt.savefig('%s_%s_splinedtrace_with_markers.png' % (save_file, channel_name))
        # plt.close()

        plt.pcolormesh(self.full_image_data, vmax=vmaxval, vmin=vminval)
        plt.colorbar()
        for dna_num in sorted(self.splined_traces.keys()):
            plt.plot(self.splined_traces[dna_num][:, 0], self.splined_traces[dna_num][:, 1], color="c", linewidth=1.0)
        # plt.savefig("%s_%s_splinedtrace.png" % (save_file, channel_name))
        plt.savefig(output_dir / filename / f"{channel_name}_splinedtrace.png")
        LOGGER.info(f"Splined Trace image saved to : {str(output_dir / filename / f'{channel_name}_splinedtrace.png')}")
        plt.close()

        """
        plt.pcolormesh(self.full_image_data)
        plt.colorbar()
        for dna_num in sorted(self.ordered_traces.keys()):
            #disordered_trace_list = self.ordered_traces[dna_num].tolist()
            #less_dense_trace = np.array([disordered_trace_list[i] for i in range(0,len(disordered_trace_list),5)])
            plt.plot(self.ordered_traces[dna_num][:,0], self.ordered_traces[dna_num][:,1])
        plt.savefig('%s_%s_splinedtrace.png' % (save_file, channel_name))
        plt.close()
        """

        plt.pcolormesh(self.full_image_data, vmax=vmaxval, vmin=vminval)
        plt.colorbar()
        for dna_num in sorted(self.disordered_trace.keys()):
            # disordered_trace_list = self.disordered_trace[dna_num].tolist()
            # less_dense_trace = np.array([disordered_trace_list[i] for i in range(0,len(disordered_trace_list),5)])
            plt.plot(
                self.disordered_trace[dna_num][:, 0],
                self.disordered_trace[dna_num][:, 1],
                "o",
                markersize=0.5,
                color="c",
            )
        # plt.savefig("%s_%s_disorderedtrace.png" % (save_file, channel_name))
        plt.savefig(output_dir / filename / f"{channel_name}_disordered_trace.png")
        plt.close()
        LOGGER.info(
            f"Disordered trace image saved to : {str(output_dir / filename / f'{channel_name}_disordered_trace.png')}"
        )

        plt.pcolormesh(self.full_image_data, vmax=vmaxval, vmin=vminval)
        plt.colorbar()
        for dna_num in sorted(self.grains.keys()):
            grain_plt = np.argwhere(self.grains[dna_num] == 1)
            plt.plot(grain_plt[:, 0], grain_plt[:, 1], "o", markersize=2, color="c")
        # plt.savefig("%s_%s_grains.png" % (save_file, channel_name))
        plt.savefig(output_dir / filename / f"{channel_name}_grains.png")
        plt.close()
        LOGGER.info(f"Grains image saved to : {str(output_dir / filename / f'{channel_name}_grains.png')}")

    # FIXME : Replace with Path() (.mkdir(parent=True, exists=True) negate need to handle errors.)
    def _checkForSaveDirectory(self, filename, new_directory_name):

        split_directory_path = os.path.split(filename)

        try:
            os.mkdir(os.path.join(split_directory_path[0], new_directory_name))
        except OSError:  # OSError happens if the directory already exists
            pass

        updated_filename = os.path.join(split_directory_path[0], new_directory_name, split_directory_path[1])

        return updated_filename

    def findWrithe(self):
        pass

    def find_curvature(self):

        # FIXME : Iterate directly over self.splined_traces.values() or self.splined_traces.items()
        for dna_num in sorted(self.splined_traces.keys()):  # the number of molecules identified
            # splined_traces is a dictionary, where the keys are the number of the molecule, and the values are a
            # list of coordinates, in a numpy.ndarray
            # if self.mol_is_circular[dna_num]:
            curve = []
            contour = 0
            coordinates = np.zeros([2, self.neighbours * 2 + 1])
            for i, (x, y) in enumerate(self.splined_traces[dna_num]):
                # Extracts the coordinates for the required number of points and puts them in an array
                if self.mol_is_circular[dna_num] or (
                    self.neighbours < i < len(self.splined_traces[dna_num]) - self.neighbours
                ):
                    for j in range(self.neighbours * 2 + 1):
                        coordinates[0][j] = self.splined_traces[dna_num][i - j][0]
                        coordinates[1][j] = self.splined_traces[dna_num][i - j][1]

                    # Calculates the angles for the tangent lines to the left and the right of the point
                    theta1 = math.atan(
                        (coordinates[1][self.neighbours] - coordinates[1][0])
                        / (coordinates[0][self.neighbours] - coordinates[0][0])
                    )
                    theta2 = math.atan(
                        (coordinates[1][-1] - coordinates[1][self.neighbours])
                        / (coordinates[0][-1] - coordinates[0][self.neighbours])
                    )

                    left = coordinates[:, : self.neighbours + 1]
                    right = coordinates[:, -(self.neighbours + 1) :]

                    xa = np.mean(left[0])
                    ya = np.mean(left[1])

                    xb = np.mean(right[0])
                    yb = np.mean(right[1])

                    # Calculates the curvature using the change in angle divided by the distance
                    dist = math.hypot((xb - xa), (yb - ya))
                    dist_real = dist * self.pixel_size
                    curve.append([i, contour, (theta2 - theta1) / dist_real])

                    contour = contour + math.hypot(
                        (coordinates[0][self.neighbours] - coordinates[0][self.neighbours - 1]),
                        (coordinates[1][self.neighbours] - coordinates[1][self.neighbours - 1]),
                    )
                self.curvature[dna_num] = curve

    def saveCurvature(self):

        # FIXME : Iterate directly over self.splined_traces.values() or self.splined_traces.items()
        # roc_array = np.zeros(shape=(1, 3))
        for dna_num in sorted(self.curvature.keys()):
            for i, [n, contour, c] in enumerate(self.curvature[dna_num]):
                try:
                    roc_array = np.append(roc_array, np.array([[dna_num, i, contour, c]]), axis=0)
                    # oc_array.append([dna_num, i, contour, c])
                except NameError:
                    roc_array = np.array([[dna_num, i, contour, c]])
                # roc_array = np.vstack((roc_array, np.array([dna_num, i, c])))
        # roc_array = np.delete(roc_array, 0, 0)
        roc_stats = pd.DataFrame(roc_array)

        if not os.path.exists(os.path.join(os.path.dirname(self.filename), "Curvature")):
            os.mkdir(os.path.join(os.path.dirname(self.filename), "Curvature"))
        directory = os.path.join(os.path.dirname(self.filename), "Curvature")
        savename = os.path.join(directory, os.path.basename(self.filename)[:-4])
        roc_stats.to_json(savename + ".json")
        roc_stats.to_csv(savename + ".csv")

    def plotCurvature(self, dna_num):

        """Plot the curvature of the chosen molecule as a function of the contour length (in metres)"""

        curvature = np.array(self.curvature[dna_num])
        length = len(curvature)
        # FIXME : Replace with Path()
        if not os.path.exists(os.path.join(os.path.dirname(self.filename), "Curvature")):
            os.mkdir(os.path.join(os.path.dirname(self.filename), "Curvature"))
        directory = os.path.join(os.path.dirname(self.filename), "Curvature")
        savename = os.path.join(directory, os.path.basename(self.filename)[:-4])

        plt.figure()
        sns.lineplot(curvature[:, 1] * self.pixel_size, curvature[:, 2], color="k")
        plt.ylim(-1e9, 1e9)
        plt.ticklabel_format(axis="both", style="sci", scilimits=(0, 0))
        plt.axvline(curvature[0][1], color="#D55E00")
        plt.axvline(curvature[int(length / 6)][1] * self.pixel_size, color="#E69F00")
        plt.axvline(curvature[int(length / 6 * 2)][1] * self.pixel_size, color="#F0E442")
        plt.axvline(curvature[int(length / 6 * 3)][1] * self.pixel_size, color="#009E74")
        plt.axvline(curvature[int(length / 6 * 4)][1] * self.pixel_size, color="#0071B2")
        plt.axvline(curvature[int(length / 6 * 5)][1] * self.pixel_size, color="#CC79A7")
        plt.savefig("%s_%s_curvature.png" % (savename, dna_num))
        plt.close()

    def measure_contour_length(self):

        """Measures the contour length for each of the splined traces taking into
        account whether the molecule is circular or linear

        Contour length units are nm"""

        for dna_num in sorted(self.splined_traces.keys()):

            if self.mol_is_circular[dna_num]:
                for num, i in enumerate(self.splined_traces[dna_num]):
                    x1 = self.splined_traces[dna_num][num - 1, 0]
                    y1 = self.splined_traces[dna_num][num - 1, 1]
                    x2 = self.splined_traces[dna_num][num, 0]
                    y2 = self.splined_traces[dna_num][num, 1]

                    try:
                        hypotenuse_array.append(math.hypot((x1 - x2), (y1 - y2)))
                    except NameError:
                        hypotenuse_array = [math.hypot((x1 - x2), (y1 - y2))]

                self.contour_lengths[dna_num] = np.sum(np.array(hypotenuse_array)) * self.pixel_size * 1e9
                del hypotenuse_array

            else:
                for num, i in enumerate(self.splined_traces[dna_num]):
                    try:
                        x1 = self.splined_traces[dna_num][num, 0]
                        y1 = self.splined_traces[dna_num][num, 1]
                        x2 = self.splined_traces[dna_num][num + 1, 0]
                        y2 = self.splined_traces[dna_num][num + 1, 1]

                        try:
                            hypotenuse_array.append(math.hypot((x1 - x2), (y1 - y2)))
                        except NameError:
                            hypotenuse_array = [math.hypot((x1 - x2), (y1 - y2))]
                    except IndexError:  # IndexError happens at last point in array
                        self.contour_lengths[dna_num] = np.sum(np.array(hypotenuse_array)) * self.pixel_size * 1e9
                        del hypotenuse_array
                        break

    def writeContourLengths(self, filename, channel_name):

        if not self.contour_lengths:
            self.measure_contour_length()

        with open(f"{filename}_{channel_name}_contours.txt", "w") as writing_file:
            writing_file.write("#units: nm\n")
            for dna_num in sorted(self.contour_lengths.keys()):
                writing_file.write("%f \n" % self.contour_lengths[dna_num])

    # FIXME : This method doesn't appear to be used here nor within pygwytracing, can it be removed?
    def writeCoordinates(self, dna_num):
        # FIXME: Replace with Path()
        if not os.path.exists(os.path.join(os.path.dirname(self.filename), "Coordinates")):
            os.mkdir(os.path.join(os.path.dirname(self.filename), "Coordinates"))
        directory = os.path.join(os.path.dirname(self.filename), "Coordinates")
        savename = os.path.join(directory, os.path.basename(self.filename)[:-4])
        for i, (x, y) in enumerate(self.splined_traces[dna_num]):
            try:
                coordinates_array = np.append(coordinates_array, np.array([[x, y]]), axis=0)
            except NameError:
                coordinates_array = np.array([[x, y]])

        coordinates = pd.DataFrame(coordinates_array)
        coordinates.to_csv("%s_%s.csv" % (savename, dna_num))

        plt.plot(coordinates_array[:, 0], coordinates_array[:, 1], "ko")
        plt.savefig("%s_%s_coordinates.png" % (savename, dna_num))

    def measure_end_to_end_distance(self):
        """Calculate the Euclidean distance between the start and end of linear molecules.
        The hypotenuse is calculated between the start ([0,0], [0,1]) and end ([-1,0], [-1,1]) of linear
        molecules. If the molecule is circular then the distance is set to zero (0).
        """

        for dna_num in sorted(self.splined_traces.keys()):
            if self.mol_is_circular[dna_num]:
                self.end_to_end_distance[dna_num] = 0
            else:
                x1 = self.splined_traces[dna_num][0, 0]
                y1 = self.splined_traces[dna_num][0, 1]
                x2 = self.splined_traces[dna_num][-1, 0]
                y2 = self.splined_traces[dna_num][-1, 1]
                self.end_to_end_distance[dna_num] = math.hypot((x1 - x2), (y1 - y2)) * self.pixel_size * 1e9
        # self.end_to_end_distance = {
        #     dna_num: math.hypot((trace[0, 0] - trace[-1, 0]), (trace[0, 1] - trace[-1, 1]))
        #     if self.mol_is_circular[dna_num]
        #     else 0
        #     for dna_num, trace in self.splined_traces.items()
        # }


class traceStats(object):
    """Combine and save trace statistics."""

    def __init__(self, trace_object: dnaTrace, image_path: Union[str, Path]) -> None:
        """Initialise the class.

        Parameters
        ----------
        trace_object: dnaTrace
            Object produced from tracing.
        image_path: Union[str, Path]
            Path for saving images to.

        Returns
        -------
        None
        """

        self.trace_object = trace_object
        self.image_path = Path(image_path)
        self.df = []
        self.create_trace_stats()

    def create_trace_stats(self):
        """Creates a pandas dataframe of the contour length, whether its circular and end to end distance
        combined with details of the working directory, directory images were found in and the image name.
        """
        stats = OrderedDict()
        for mol_num, _ in self.trace_object.ordered_traces.items():
            stats[mol_num] = {}
            stats[mol_num]["Contour Lengths"] = self.trace_object.contour_lengths[mol_num]
            stats[mol_num]["Circular"] = self.trace_object.mol_is_circular[mol_num]
            stats[mol_num]["End to End Distance"] = self.trace_object.end_to_end_distance[mol_num]
        self.df = pd.DataFrame.from_dict(data=stats, orient="index")
        self.df.reset_index(drop=True, inplace=True)
        self.df.index.name = "Molecule Number"
        # self.df["Experiment Directory"] = str(Path().cwd())
        self.df["Image Name"] = self.image_path.name
        self.df["Basename"] = str(self.image_path.parent)

    def save_trace_stats(self, save_path: Union[str, Path], json: bool = True, csv: bool = True) -> None:
        """Write trace statistics to JSON and/or CSV.

        Parameters
        ----------
        save_path: Union[str, Path]
            Directory to save results to.
        json: bool
            Whether to save a JSON version of statistics.
        csv: bool
            Whether to save a CSV version of statistics.
        """
        if json:
            self.df.to_json(save_path / "tracestats.json")
            LOGGER.info(f"Saved trace info for all analysed images to: {str(save_path / 'tracestats.json')}")
        if csv:
            self.df.to_csv(save_path / "tracestats.csv")
            LOGGER.info(f"Saved trace info for all analysed images to: {str(save_path / 'tracestats.csv')}")



"""Skeletonize molecules"""
import logging
from typing import Callable
import numpy as np
from skimage.morphology import medial_axis, skeletonize, thin

from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)


def get_skeleton(image: np.ndarray, method: str) -> np.ndarray:
    """Factory method for skeletonizing molecules.

    Parameters
    ----------

    image : np.ndarray
        Image of molecule to be skeletonized.

    method : str
        Method to use, default is 'zhang' other options are 'lee', 'medial_axis' and 'thin'.

    Returns
    -------
    np.ndarray
        Skeletonised version of the image.all($0)

    Notes
    -----

    This is a thin wrapper to the methods provided
    by the `skimage.morphology
    <https://scikit-image.org/docs/stable/api/skimage.morphology.html?highlight=skeletonize>`_
    module. See also the `examples
    <https://scikit-image.org/docs/stable/auto_examples/edges/plot_skeleton.html>_
    """
    skeletonizer = _get_skeletonize(method)
    return skeletonizer(image)


def _get_skeletonize(method: str = "zhang") -> Callable:
    """Creator component which determines which skeletonize method to use.

    Parameters
    ----------
    method: str
        Method to use for skeletonizing, methods are 'zhang' (default), 'lee', 'medial_axis', and 'thin'.

    Returns
    -------
    Callable
        Returns the function appropriate for the required skeletonizing method.
    """
    if method == "zhang":
        return _skeletonize_zhang
    if method == "lee":
        return _skeletonize_lee
    if method == "medial_axis":
        return _skeletonize_medial_axis
    if method == "thin":
        return _skeletonize_thin
    raise ValueError(method)


def _skeletonize_zhang(image: np.ndarray) -> np.ndarray:
    return skeletonize(image, method="zhang")


def _skeletonize_lee(image: np.ndarray) -> np.ndarray:
    return skeletonize(image, method="lee")


def _skeletonize_medial_axis(image: np.ndarray) -> np.ndarray:
    return medial_axis(image, return_distance=False)


def _skeletonize_thin(image: np.ndarray) -> np.ndarray:
    return thin(image)



"""Utilities"""
from argparse import Namespace
import logging
from pathlib import Path
from typing import Union, List, Dict
from collections import defaultdict

import numpy as np
import pandas as pd

from topostats.thresholds import threshold
from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)


ALL_STATISTICS_COLUMNS = (
    "Molecule Number",
    "centre_x",
    "centre_y",
    "radius_min",
    "radius_max",
    "radius_mean",
    "radius_median",
    "height_min",
    "height_max",
    "height_median",
    "height_mean",
    "volume",
    "area",
    "area_cartesian_bbox",
    "smallest_bounding_width",
    "smallest_bounding_length",
    "smallest_bounding_area",
    "aspect_ratio",
    "Contour Lengths",
    "Circular",
    "End to End Distance",
    "Image Name",
    "Basename",
)


def convert_path(path: Union[str, Path]) -> Path:
    """Ensure path is Path object.

    Parameters
    ----------
    path: Union[str, Path]
        Path to be converted.

    Returns
    -------
    Path
        pathlib Path
    """
    return Path().cwd() if path == "./" else Path(path).expanduser()


def find_images(base_dir: Union[str, Path] = None, file_ext: str = ".spm") -> List:
    """Scan the specified directory for images with the given file extension.

    Parameters
    ----------
    base_dir: Union[str, Path]
        Directory to recursively search for files, if not specified the current directory is scanned.
    file_ext: str
        File extension to search for.

    Returns
    -------
    List
        List of files found with the extension in the given directory.
    """
    base_dir = Path("./") if base_dir is None else Path(base_dir)
    return list(base_dir.glob("**/*" + file_ext))


def get_out_path(
    image_path: Union[str, Path] = None, base_dir: Union[str, Path] = None, output_dir: Union[str, Path] = None
) -> Path:
    """Adds the image path relative to the base directory to the output directory.

    Parameters
    ----------
    image_path: Path
        The path of the current image.
    base_dir: Path
        Directory to recursively search for files.
    output_dir: Path
        The output directory specified in the configuration file.

    Returns
    -------
    Path
        The output path that mirrors the input path structure.
    """
    # If image_path is relative and doesn't include base_dir then a ValueError is raisedin which
    # case we just want to append the image_path to the output_dir
    try:
        # Remove the filename if there is a suffix, not always the case as
        # get_out_path is called from folder_grainstats()
        if image_path.suffix:
            return output_dir / image_path.parent.relative_to(base_dir)
        else:
            return output_dir / image_path.relative_to(base_dir)
    except ValueError:
        if image_path.suffix:
            return output_dir / image_path.parent
        else:
            return output_dir / image_path
    # AttributeError is raised if image_path is a string (since it isn't a Path() object with a .suffix)
    except AttributeError:
        LOGGER.error("A string form of a Path has been passed to 'get_out_path()' for image_path")
        raise


def update_config(config: dict, args: Union[dict, Namespace]) -> Dict:
    """Update the configuration with any arguments

    Parameters
    ----------
    config: dict
        Dictionary of configuration (typically read from YAML file specified with '-c/--config <filename>')
    args: Namespace
        Command line arguments
    Returns
    -------
    Dict
        Dictionary updated with command arguments.
    """
    args = vars(args) if isinstance(args, Namespace) else args

    config_keys = config.keys()
    for arg_key, arg_value in args.items():
        if isinstance(arg_value, dict):
            update_config(config, arg_value)
        else:
            if arg_key in config_keys and arg_value is not None:
                original_value = config[arg_key]
                config[arg_key] = arg_value
                LOGGER.info(f"Updated config config[{arg_key}] : {original_value} > {arg_value} ")
    config["base_dir"] = convert_path(config["base_dir"])
    config["output_dir"] = convert_path(config["output_dir"])
    return config


def _get_mask(image: np.ndarray, threshold: float, threshold_direction: str, img_name: str = None) -> np.ndarray:
    """Calculate a mask for pixels that exceed the threshold

    Parameters
    ----------
    image: np.array
        Numpy array representing image.
    threshold: float
        A float representing the threshold
    threshold_direction: str
        A string representing the direction that should be thresholded. ("above", "below")
    img_name: str
        Name of image being processed

    Returns
    -------
    np.array
        Numpy array of image with objects coloured.
    """
    if threshold_direction == "upper":
        LOGGER.info(f"[{img_name}] : Masking (upper) Threshold: {threshold}")
        return image > threshold
    if threshold_direction == "lower":
        LOGGER.info(f"[{img_name}] : Masking (lower) Threshold: {threshold}")
        return image < threshold
    # LOGGER.fatal(f"[{img_name}] : Threshold direction invalid: {threshold_direction}")


def get_mask(image: np.ndarray, thresholds: dict, img_name: str = None) -> np.ndarray:
    """Mask data that should not be included in flattening.

    Parameters
    ----------
    image: np.ndarray
        2D Numpy array of the image to have a mask derived for.

    thresholds: dict
        Dictionary of thresholds, at a bare minimum must have key 'lower' with an associated value, second key is
        to have an 'upper' threshold.
    img_name: str
        Image name that is being masked.

    Returns
    -------
    np.ndarray
        2D Numpy boolean array of points to mask.
    """
    # Both thresholds are applicable
    if "lower" in thresholds and "upper" in thresholds:
        mask_upper = _get_mask(image, threshold=thresholds["upper"], threshold_direction="upper", img_name=img_name)
        mask_lower = _get_mask(image, threshold=thresholds["lower"], threshold_direction="lower", img_name=img_name)
        # Masks are combined to remove both the extreme high and extreme low data points.
        return mask_upper + mask_lower
    # Only lower threshold is applicable
    if "lower" in thresholds:
        return _get_mask(image, threshold=thresholds["lower"], threshold_direction="lower", img_name=img_name)
    # Only upper threshold is applicable
    return _get_mask(image, threshold=thresholds["upper"], threshold_direction="upper", img_name=img_name)


def get_thresholds(
    image: np.ndarray,
    threshold_method: str,
    otsu_threshold_multiplier: float = None,
    threshold_std_dev: dict = None,
    absolute: dict = None,
    **kwargs,
) -> Dict:
    """Obtain thresholds for masking data points.

    Parameters
    ----------
    image : np.ndarray
        2D Numpy array of image to be masked
    threshold_method : str
        Method for thresholding, 'otsu', 'std_dev' or 'absolute' are valid options.
    threshold_std_dev : dict
        Dict of upper and lower thresholds for the standard deviation method.
    absolute : tuple
        Dict of lower and upper thresholds.
    **kwargs:

    Returns
    -------
    Dict
        Dictionary of thresholds, contains keys 'lower' and optionally 'upper'.
    """
    thresholds = defaultdict()
    if threshold_method == "otsu":
        thresholds["upper"] = threshold(image, method="otsu", otsu_threshold_multiplier=otsu_threshold_multiplier)
    elif threshold_method == "std_dev":
        try:
            if threshold_std_dev["lower"] is not None:
                thresholds["lower"] = threshold(image, method="mean") - threshold_std_dev["lower"] * np.nanstd(image)
            if threshold_std_dev["upper"] is not None:
                thresholds["upper"] = threshold(image, method="mean") + threshold_std_dev["upper"] * np.nanstd(image)
        except TypeError as typeerror:
            raise typeerror
    elif threshold_method == "absolute":
        if absolute["lower"] is not None:
            thresholds["lower"] = absolute["lower"]
        if absolute["upper"] is not None:
            thresholds["upper"] = absolute["upper"]
    else:
        if not isinstance(threshold_method, str):
            raise TypeError(
                f"threshold_method ({threshold_method}) should be a string. Valid values : 'otsu' 'std_dev' 'absolute'"
            )
        if threshold_method not in ["otsu", "std_dev", "absolute"]:
            raise ValueError(
                f"threshold_method ({threshold_method}) is invalid. Valid values : 'otsu' 'std_dev' 'absolute'"
            )
    return thresholds


def create_empty_dataframe(columns: set = ALL_STATISTICS_COLUMNS) -> pd.DataFrame:
    """Create an empty data frame for returning when no results are found.

    Parameters
    ----------
    columns: list
        Columns of the empty dataframe.

    Returns
    -------
    pd.DataFrame
        Empty Pandas DataFrame.
    """
    return pd.DataFrame([np.repeat(np.nan, len(columns))], columns=columns)


def folder_grainstats(output_dir: Union[str, Path], base_dir: Union[str, Path], all_stats_df: pd.DataFrame) -> None:
    """Creates saves a data frame of grain and tracing statictics at the folder level.

    Parameters
    ----------
    output_dir: Union[str, Path]
        Path of the output directory head.
    base_dir: Union[str, Path]
        Path of the base directory where files were found.
    all_stats_df: pd.DataFrame
        The dataframe containing all sample statistics run.

    Returns
    -------
    None
        This only saves the dataframes and does not retain them.
    """
    dirs = set(all_stats_df["Basename"].values)
    try:
        for _dir in dirs:
            out_path = get_out_path(Path(_dir), base_dir, output_dir)
            all_stats_df[all_stats_df["Basename"] == _dir].to_csv(out_path / "Processed" / "folder_grainstats.csv")
            LOGGER.info(f"Folder-wise statistics saved to: {str(out_path)}/Processed/folder_grainstats.csv")
    except TypeError:
        LOGGER.info("Unable to generate folderwise statistics as 'all_statis_df' is empty")



import numpy as np
import matplotlib.pyplot as plt
import math


class getSkeleton(object):

    """Skeltonisation algorithm based on the paper "A Fast Parallel Algorithm for
    Thinning Digital Patterns" by Zhang et al., 1984"""

    def __init__(self, image_data, binary_map, number_of_columns, number_of_rows, pixel_size):
        self.image_data = image_data
        self.binary_map = binary_map
        self.number_of_columns = number_of_columns
        self.number_of_rows = number_of_rows
        self.pixel_size = pixel_size

        self.p2 = 0
        self.p3 = 0
        self.p4 = 0
        self.p5 = 0
        self.p6 = 0
        self.p7 = 0
        self.p8 = 0

        # skeletonising variables
        self.mask_being_skeletonised = []
        self.output_skeleton = []
        self.skeleton_converged = False
        self.pruning = True

        # Height checking variables
        self.average_height = 0
        # self.cropping_dict = self._initialiseHeightFindingDict()
        self.highest_points = {}
        self.search_window = int(3 / (pixel_size * 1e9))
        # Check that the search window is bigger than 0:
        if self.search_window < 2:
            self.search_window = 3
        self.dir_search = int(0.75 / (pixel_size * 1e9))
        if self.dir_search < 3:
            self.dir_search = 3

        self.getDNAmolHeightStats()
        self.doSkeletonising()

    def getDNAmolHeightStats(self):
        # Why are axes swapped here?
        self.image_data = np.swapaxes(self.image_data, 0, 1)

        # This doesn't appear to be used within this class and its not used anywhere in dnatracing.py either.
        self.average_height = np.average(self.image_data[np.argwhere(self.binary_map == 1)])
        # print(self.average_height)

    def doSkeletonising(self):

        """Simple while loop to check if the skeletonising is finished"""

        self.mask_being_skeletonised = self.binary_map

        while not self.skeleton_converged:
            self._doSkeletonisingIteration()

        # When skeleton converged do an additional iteration of thinning to remove hanging points
        self.finalSkeletonisationIteration()

        self.pruning = True
        while self.pruning:
            self.pruneSkeleton()

        self.output_skeleton = np.argwhere(self.mask_being_skeletonised == 1)

    def _doSkeletonisingIteration(self):

        """Do an iteration of skeletonisation - check for the local binary pixel
        environment and assess the local height values to decide whether to
        delete a point
        """

        number_of_deleted_points = 0
        pixels_to_delete = []

        # Sub-iteration 1 - binary check
        mask_coordinates = np.argwhere(self.mask_being_skeletonised == 1).tolist()
        for point in mask_coordinates:
            if self._deletePixelSubit1(point):
                pixels_to_delete.append(point)

        # Check the local height values to determine if pixels should be deleted
        # pixels_to_delete = self._checkHeights(pixels_to_delete)

        for x, y in pixels_to_delete:
            number_of_deleted_points += 1
            self.mask_being_skeletonised[x, y] = 0
        pixels_to_delete = []

        # Sub-iteration 2 - binary check
        mask_coordinates = np.argwhere(self.mask_being_skeletonised == 1).tolist()
        for point in mask_coordinates:
            if self._deletePixelSubit2(point):
                pixels_to_delete.append(point)

        # Check the local height values to determine if pixels should be deleted
        # pixels_to_delete = self._checkHeights(pixels_to_delete)

        for x, y in pixels_to_delete:
            number_of_deleted_points += 1
            self.mask_being_skeletonised[x, y] = 0

        if number_of_deleted_points == 0:
            self.skeleton_converged = True

    def _deletePixelSubit1(self, point):

        """Function to check whether a single point should be deleted based
        on both its local binary environment and its local height values"""

        self.p2, self.p3, self.p4, self.p5, self.p6, self.p7, self.p8, self.p9 = genTracingFuncs.getLocalPixelsBinary(
            self.mask_being_skeletonised, point[0], point[1]
        )

        if (
            self._binaryThinCheck_a()
            and self._binaryThinCheck_b()
            and self._binaryThinCheck_c()
            and self._binaryThinCheck_d()
        ):
            return True
        else:
            return False

    def _deletePixelSubit2(self, point):

        """Function to check whether a single point should be deleted based
        on both its local binary environment and its local height values"""

        self.p2, self.p3, self.p4, self.p5, self.p6, self.p7, self.p8, self.p9 = genTracingFuncs.getLocalPixelsBinary(
            self.mask_being_skeletonised, point[0], point[1]
        )

        # Add in generic code here to protect high points from being deleted
        if (
            self._binaryThinCheck_a()
            and self._binaryThinCheck_b()
            and self._binaryThinCheck_csharp()
            and self._binaryThinCheck_dsharp()
        ):
            return True
        else:
            return False

    """These functions are ripped from the Zhang et al. paper and do the basic
    skeletonisation steps

    I can use the information from the c,d,c' and d' tests to determine a good
    direction to search for higher height values """

    def _binaryThinCheck_a(self):
        # Condition A protects the endpoints (which will be > 2) - add in code here to prune low height points
        if 2 <= self.p2 + self.p3 + self.p4 + self.p5 + self.p6 + self.p7 + self.p8 + self.p9 <= 6:
            return True
        else:
            return False

    def _binaryThinCheck_b(self):
        count = 0

        if [self.p2, self.p3] == [0, 1]:
            count += 1
        if [self.p3, self.p4] == [0, 1]:
            count += 1
        if [self.p4, self.p5] == [0, 1]:
            count += 1
        if [self.p5, self.p6] == [0, 1]:
            count += 1
        if [self.p6, self.p7] == [0, 1]:
            count += 1
        if [self.p7, self.p8] == [0, 1]:
            count += 1
        if [self.p8, self.p9] == [0, 1]:
            count += 1
        if [self.p9, self.p2] == [0, 1]:
            count += 1

        if count == 1:
            return True
        else:
            return False

    def _binaryThinCheck_c(self):
        if self.p2 * self.p4 * self.p6 == 0:
            return True
        else:
            return False

    def _binaryThinCheck_d(self):
        if self.p4 * self.p6 * self.p8 == 0:
            return True
        else:
            return False

    def _binaryThinCheck_csharp(self):
        if self.p2 * self.p4 * self.p8 == 0:
            return True
        else:
            return False

    def _binaryThinCheck_dsharp(self):
        if self.p2 * self.p6 * self.p8 == 0:
            return True
        else:
            return False

    def _checkHeights(self, candidate_points):

        try:
            candidate_points = candidate_points.tolist()
        except AttributeError:
            pass

        for x, y in candidate_points:

            # if point is basically at background don't bother assessing height and just delete:
            if self.image_data[x, y] < 1e-9:
                continue

            # Check if the point has already been identified as a high point
            try:
                self.highest_points[(x, y)]
                candidate_points.pop(candidate_points.index([x, y]))
                # print(x,y)
                continue
            except KeyError:
                pass

            (
                self.p2,
                self.p3,
                self.p4,
                self.p5,
                self.p6,
                self.p7,
                self.p8,
                self.p9,
            ) = genTracingFuncs.getLocalPixelsBinary(self.mask_being_skeletonised, x, y)

            print([self.p9, self.p2, self.p3], [self.p8, 1, self.p4], [self.p7, self.p6, self.p5])

            height_points_to_check = self._checkWhichHeightPoints()
            height_points = np.around(self.cropping_dict[height_points_to_check](x, y), decimals=11)
            test_value = np.around(self.image_data[x, y], decimals=11)
            # print(height_points_to_check, [x,y], self.image_data[x,y], height_points)

            # if the candidate points is the highest local point don't delete it
            if test_value >= sorted(height_points)[-1]:
                print([self.p9, self.p2, self.p3], [self.p8, 1, self.p4], [self.p7, self.p6, self.p5])
                print(height_points_to_check, [x, y], self.image_data[x, y], height_points)
                self.highest_points[(x, y)] = height_points_to_check
                candidate_points.pop(candidate_points.index([x, y]))
                print(height_points_to_check, (x, y))
            else:
                x_n, y_n = self._identifyHighestPoint(x, y, height_points_to_check, height_points)
                self.highest_points[(x_n, y_n)] = height_points_to_check
                pass

        return candidate_points

    def _checkWhichHeightPoints(self):

        # Is the point on the left hand edge?
        # if (self.p8 == 1 and self.p4 == 0 and self.p2 == self.p6):
        if self.p7 + self.p8 + self.p9 == 3 and self.p3 + self.p4 + self.p5 == 0 and self.p2 == self.p6:
            """e.g. [1, 1, 0]
            [1, 1, 0]
            [1, 1, 0]"""
            return "horiz_left"
        # elif (self.p8 == 0 and self.p4 == 1 and self.p2 == self.p6):
        elif self.p7 + self.p8 + self.p9 == 0 and self.p3 + self.p4 + self.p5 == 3 and self.p2 == self.p6:
            """e.g. [0, 1, 1]
            [0, 1, 1]
            [0, 1, 1]"""
            return "horiz_right"
        # elif (self.p2 == 1 and self.p6 == 0 and self.p4 == self.p8):
        elif self.p9 + self.p2 + self.p3 == 3 and self.p5 + self.p6 + self.p7 == 0 and self.p4 == self.p8:
            """e.g. [1, 1, 1]
            [1, 1, 1]
            [0, 0, 0]"""
            return "vert_up"
        # elif (self.p2 == 0 and self.p6 == 1 and self.p4 == self.p8):
        elif (
            self.p9 + self.p2 + self.p3 == 0 and self.p5 + self.p6 + self.p7 == 3 and self.p4 == self.p8
        ):  # and self.p4 == self.p8):
            """e.g. [0, 0, 0]
            [1, 1, 1]
            [1, 1, 1]"""
            return "vert_down"
        elif self.p2 + self.p8 <= 1 and self.p4 + self.p5 + self.p6 >= 2:
            """e.g. [0, 0, 1]       [0, 0, 0]
            [0, 1, 1]       [0, 1, 1]
            [1, 1, 1]   or  [0, 1, 1]"""
            return "diagright_down"
        elif self.p4 + self.p6 <= 1 and self.p8 + self.p9 + self.p2 >= 2:
            """e.g. [1, 1, 1]       [1, 1, 0]
            [1, 1, 0]       [1, 1, 0]
            [1, 0, 0]   or  [0, 0, 0]"""
            return "diagright_up"
        elif self.p2 + self.p4 <= 1 and self.p8 + self.p7 + self.p6 >= 2:
            """e.g. [1, 0, 0]       [0, 0, 0]
            [1, 1, 0]       [1, 1, 0]
            [1, 1, 1]   or  [1, 1, 0]"""
            return "diagleft_down"
        elif self.p8 + self.p6 <= 1 and self.p2 + self.p3 + self.p4 >= 2:
            """e.g. [1, 1, 1]       [0, 1, 1]
            [0, 1, 1]       [0, 1, 1]
            [0, 0, 1]   or  [0, 0, 0]"""
            return "diagleft_up"
        # else:
        #    return 'save'

    def _initialiseHeightFindingDict(self):
        height_cropping_funcs = {}

        height_cropping_funcs["horiz_left"] = self._getHorizontalLeftHeights
        height_cropping_funcs["horiz_right"] = self._getHorizontalRightHeights
        height_cropping_funcs["vert_up"] = self._getVerticalUpwardHeights
        height_cropping_funcs["vert_down"] = self._getVerticalDonwardHeights
        height_cropping_funcs["diagleft_up"] = self._getDiaganolLeftUpwardHeights
        height_cropping_funcs["diagleft_down"] = self._getDiaganolLeftDownwardHeights
        height_cropping_funcs["diagright_up"] = self._getHorizontalRightHeights
        height_cropping_funcs["diagright_down"] = self._getHorizontalRightHeights
        height_cropping_funcs["save"] = self._savePoint

        return height_cropping_funcs

    def _getHorizontalLeftHeights(self, x, y):
        heights = []  # [self.image_data[x,y]]

        for i in range(-self.search_window, self.search_window):
            if i == 0:
                continue
            heights.append(self.image_data[x - i, y])
        return heights

    def _getHorizontalRightHeights(self, x, y):
        heights = []  # [self.image_data[x,y]]

        for i in range(-self.search_window, self.search_window):
            if i == 0:
                continue
            heights.append(self.image_data[x + i, y])
        return heights

    def _getVerticalUpwardHeights(self, x, y):
        heights = []  # [self.image_data[x,y]]

        for i in range(-self.search_window, self.search_window):
            if i == 0:
                continue
            heights.append(self.image_data[x, y + i])
        return heights

    def _getVerticalDonwardHeights(self, x, y):
        heights = []  # [self.image_data[x,y]]

        for i in range(-self.search_window, self.search_window):
            if i == 0:
                continue
            heights.append(self.image_data[x, y - i])
        return heights

    def _getDiaganolLeftUpwardHeights(self, x, y):
        heights = []  # [self.image_data[x,y]]

        for i in range(-self.search_window, self.search_window):
            if i == 0:
                continue
            heights.append(self.image_data[x + i, y + i])
        return heights

    def _getDiaganolLeftDownwardHeights(self, x, y):
        heights = []  # [self.image_data[x,y]]

        for i in range(-self.search_window, self.search_window):
            if i == 0:
                continue
            heights.append(self.image_data[x - i, y - i])
        return heights

    def _getDiaganolRightUpwardHeights(self, x, y):
        heights = []  # [self.image_data[x,y]]

        for i in range(-self.search_window, self.search_window):
            if i == 0:
                continue
            heights.append(self.image_data[x - i, y + i])
        return heights

    def _getDiaganolRightDownwardHeights(self, x, y):
        heights = []  # [self.image_data[x,y]]

        for i in range(-self.search_window, self.search_window):
            if i == 0:
                continue
            heights.append(self.image_data[x + i, y - i])
        return heights

    def _condemnPoint(self, x, y):
        heights = []  # [self.image_data[x,y]]

        for i in range(1, self.search_window):
            heights.append(10)
        return heights

    def _identifyHighestPoint(self, x, y, index_direction, indexed_heights):
        highest_value = 0

        offset = len(indexed_heights) / 2

        for num, height_value in enumerate(indexed_heights):
            if height_value > highest_value:
                highest_point = height_value
                index_position = (num + 1) - offset

        if index_direction == "horiz_left":
            return x - num, y
        elif index_direction == "horiz_right":
            return x + num, y
        elif index_direction == "vert_up":
            return x, y + num
        elif index_direction == "vert_down":
            return x, y - num
        elif index_direction == "diagleft_up":
            return x + num, y + num
        elif index_direction == "diagleft_down":
            return x + num, y - num
        elif index_direction == "diagright_up":
            return x - num, y + num
        elif index_direction == "diagright_down":
            return x - num, y - num

    def finalSkeletonisationIteration(self):

        """A final skeletonisation iteration that removes "hanging" pixels.
        Examples of such pixels are:

                    [0, 0, 0]               [0, 1, 0]            [0, 0, 0]
                    [0, 1, 1]               [0, 1, 1]            [0, 1, 1]
            case 1: [0, 1, 0]   or  case 2: [0, 1, 0] or case 3: [1, 1, 0]

        This is useful for the future functions that rely on local pixel environment
        to make assessments about the overall shape/structure of traces"""

        remaining_coordinates = np.argwhere(self.mask_being_skeletonised).tolist()

        for x, y in remaining_coordinates:

            (
                self.p2,
                self.p3,
                self.p4,
                self.p5,
                self.p6,
                self.p7,
                self.p8,
                self.p9,
            ) = genTracingFuncs.getLocalPixelsBinary(self.mask_being_skeletonised, x, y)

            # Checks for case 1 pixels
            if self._binaryThinCheck_b_returncount() == 2 and self._binaryFinalThinCheck_a():
                self.mask_being_skeletonised[x, y] = 0
            # Checks for case 2 pixels
            elif self._binaryThinCheck_b_returncount() == 3 and self._binaryFinalThinCheck_b():
                self.mask_being_skeletonised[x, y] = 0

    def _binaryFinalThinCheck_a(self):

        if self.p2 * self.p4 == 1:
            return True
        elif self.p4 * self.p6 == 1:
            return True
        elif self.p6 * self.p8 == 1:
            return True
        elif self.p8 * self.p2 == 1:
            return True

    def _binaryFinalThinCheck_b(self):

        if self.p2 * self.p4 * self.p6 == 1:
            return True
        elif self.p4 * self.p6 * self.p8 == 1:
            return True
        elif self.p6 * self.p8 * self.p2 == 1:
            return True
        elif self.p8 * self.p2 * self.p4 == 1:
            return True

    def _binaryThinCheck_b_returncount(self):
        count = 0

        if [self.p2, self.p3] == [0, 1]:
            count += 1
        if [self.p3, self.p4] == [0, 1]:
            count += 1
        if [self.p4, self.p5] == [0, 1]:
            count += 1
        if [self.p5, self.p6] == [0, 1]:
            count += 1
        if [self.p6, self.p7] == [0, 1]:
            count += 1
        if [self.p7, self.p8] == [0, 1]:
            count += 1
        if [self.p8, self.p9] == [0, 1]:
            count += 1
        if [self.p9, self.p2] == [0, 1]:
            count += 1

        return count

    def pruneSkeleton(self):

        """Function to remove the hanging branches from the skeletons - these
        are a persistent problem in the overall tracing process."""

        number_of_branches = 0
        coordinates = np.argwhere(self.mask_being_skeletonised == 1).tolist()

        # The branches are typically short so if a branch is longer than a quarter
        # of the total points its assumed to be part of the real data
        length_of_trace = len(coordinates)
        max_branch_length = int(length_of_trace * 0.15)

        # _deleteSquareEnds(coordinates)

        # first check to find all the end coordinates in the trace
        potential_branch_ends = self._findBranchEnds(coordinates)

        # Now check if its a branch - and if it is delete it
        for x_b, y_b in potential_branch_ends:
            branch_coordinates = [[x_b, y_b]]
            branch_continues = True
            temp_coordinates = coordinates[:]
            temp_coordinates.pop(temp_coordinates.index([x_b, y_b]))

            count = 0

            while branch_continues:
                no_of_neighbours, neighbours = genTracingFuncs.countandGetNeighbours(x_b, y_b, temp_coordinates)

                # If branch continues
                if no_of_neighbours == 1:
                    x_b, y_b = neighbours[0]
                    branch_coordinates.append([x_b, y_b])
                    temp_coordinates.pop(temp_coordinates.index([x_b, y_b]))

                # If the branch reaches the edge of the main trace
                elif no_of_neighbours > 1:
                    branch_coordinates.pop(branch_coordinates.index([x_b, y_b]))
                    branch_continues = False
                    is_branch = True
                # Weird case that happens sometimes
                elif no_of_neighbours == 0:
                    is_branch = True
                    branch_continues = False

                if len(branch_coordinates) > max_branch_length:
                    branch_continues = False
                    is_branch = False

            if is_branch:
                number_of_branches += 1
                for x, y in branch_coordinates:
                    self.mask_being_skeletonised[x, y] = 0

        remaining_coordinates = np.argwhere(self.mask_being_skeletonised)

        if number_of_branches == 0:
            self.pruning = False

    def _findBranchEnds(self, coordinates):
        potential_branch_ends = []

        # Most of the branch ends are just points with one neighbour
        for x, y in coordinates:
            if genTracingFuncs.countNeighbours(x, y, coordinates) == 1:
                potential_branch_ends.append([x, y])
        # Find the ends that are 3/4 neighbouring points
        return potential_branch_ends

    def _deleteSquareEnds(self, coordinates):

        for x, y in coordinates:
            pass


class reorderTrace:
    @staticmethod
    def linearTrace(trace_coordinates):

        """My own function to order the points from a linear trace.

        This works by checking the local neighbours for a given pixel (starting
        at one of the ends). If this pixel has only one neighbour in the array
        of unordered points, this must be the next pixel in the trace -- and it
        is added to the ordered points trace and removed from the
        remaining_unordered_coords array.

        If there is more than one neighbouring pixel, a fairly simple function
        (checkVectorsCandidatePoints) finds which pixel incurs the smallest
        change in angle compared with the rest of the trace and chooses that as
        the next point.

        This process is repeated until all the points are placed in the ordered
        trace array or the other end point is reached."""

        try:
            trace_coordinates = trace_coordinates.tolist()
        except AttributeError:  # array is already a python list
            pass

        # Find one of the end points
        for i, (x, y) in enumerate(trace_coordinates):
            if genTracingFuncs.countNeighbours(x, y, trace_coordinates) == 1:
                ordered_points = [[x, y]]
                trace_coordinates.pop(i)
                break

        remaining_unordered_coords = trace_coordinates[:]

        while remaining_unordered_coords:

            if len(ordered_points) > len(trace_coordinates):
                break

            x_n, y_n = ordered_points[-1]  # get the last point to be added to the array and find its neighbour

            no_of_neighbours, neighbour_array = genTracingFuncs.countandGetNeighbours(
                x_n, y_n, remaining_unordered_coords
            )

            if (
                no_of_neighbours == 1
            ):  # if there's only one candidate - its the next point add it to array and delete from candidate points
                ordered_points.append(neighbour_array[0])
                remaining_unordered_coords.pop(remaining_unordered_coords.index(neighbour_array[0]))
                continue
            elif no_of_neighbours > 1:
                best_next_pixel = genTracingFuncs.checkVectorsCandidatePoints(x_n, y_n, ordered_points, neighbour_array)
                ordered_points.append(best_next_pixel)
                remaining_unordered_coords.pop(remaining_unordered_coords.index(best_next_pixel))
                continue
            elif no_of_neighbours == 0:
                # nn, neighbour_array_all_coords = genTracingFuncs.countandGetNeighbours(x_n, y_n, trace_coordinates)
                # best_next_pixel = genTracingFuncs.checkVectorsCandidatePoints(x_n, y_n, ordered_points, neighbour_array_all_coords)
                best_next_pixel = genTracingFuncs.findBestNextPoint(
                    x_n, y_n, ordered_points, remaining_unordered_coords
                )

                if not best_next_pixel:
                    return np.array(ordered_points)

                ordered_points.append(best_next_pixel)

            # If the tracing has reached the other end of the trace then its finished
            if genTracingFuncs.countNeighbours(x_n, y_n, trace_coordinates) == 1:
                break

        return np.array(ordered_points)

    @staticmethod
    def circularTrace(trace_coordinates):

        """An alternative implementation of the linear tracing algorithm but
        with some adaptations to work with circular dna molecules"""

        try:
            trace_coordinates = trace_coordinates.tolist()
        except AttributeError:  # array is already a python list
            pass

        remaining_unordered_coords = trace_coordinates[:]

        # Find a sensible point to start of the end points
        for i, (x, y) in enumerate(trace_coordinates):
            if genTracingFuncs.countNeighbours(x, y, trace_coordinates) == 2:
                ordered_points = [[x, y]]
                remaining_unordered_coords.pop(i)
                break

        # Randomly choose one of the neighbouring points as the next point
        x_n = ordered_points[0][0]
        y_n = ordered_points[0][1]
        no_of_neighbours, neighbour_array = genTracingFuncs.countandGetNeighbours(x_n, y_n, remaining_unordered_coords)
        ordered_points.append(neighbour_array[0])
        remaining_unordered_coords.pop(remaining_unordered_coords.index(neighbour_array[0]))

        count = 0

        while remaining_unordered_coords:

            x_n, y_n = ordered_points[-1]  # get the last point to be added to the array and find its neighbour

            no_of_neighbours, neighbour_array = genTracingFuncs.countandGetNeighbours(
                x_n, y_n, remaining_unordered_coords
            )

            if (
                no_of_neighbours == 1
            ):  # if there's only one candidate - its the next point add it to array and delete from candidate points
                ordered_points.append(neighbour_array[0])
                remaining_unordered_coords.pop(remaining_unordered_coords.index(neighbour_array[0]))
                continue

            elif no_of_neighbours > 1:
                best_next_pixel = genTracingFuncs.checkVectorsCandidatePoints(x_n, y_n, ordered_points, neighbour_array)
                ordered_points.append(best_next_pixel)
                remaining_unordered_coords.pop(remaining_unordered_coords.index(best_next_pixel))
                continue

            elif len(ordered_points) > len(trace_coordinates):
                vector_start_end = abs(
                    math.hypot(
                        ordered_points[0][0] - ordered_points[-1][0], ordered_points[0][1] - ordered_points[-1][1]
                    )
                )
                if vector_start_end > 5:  # Checks if trace has basically finished i.e. is close to where it started
                    ordered_points.pop(-1)
                    return np.array(ordered_points), False
                else:
                    break

            elif no_of_neighbours == 0:
                # Check if the tracing is finished
                nn, neighbour_array_all_coords = genTracingFuncs.countandGetNeighbours(x_n, y_n, trace_coordinates)
                if ordered_points[0] in neighbour_array_all_coords:
                    break

                    # Checks for bug that happens when tracing messes up
                if ordered_points[-1] == ordered_points[-3]:
                    ordered_points = ordered_points[:-6]
                    return np.array(ordered_points), False

                # Maybe at a crossing with all neighbours deleted - this is crucially a point where errors often occur
                else:
                    # best_next_pixel = genTracingFuncs.checkVectorsCandidatePoints(x_n, y_n, ordered_points, remaining_unordered_coords)
                    best_next_pixel = genTracingFuncs.findBestNextPoint(
                        x_n, y_n, ordered_points, remaining_unordered_coords
                    )

                    if not best_next_pixel:
                        return np.array(ordered_points), False

                    vector_to_new_point = abs(math.hypot(best_next_pixel[0] - x_n, best_next_pixel[1] - y_n))

                    if vector_to_new_point > 5:  # arbitary distinction but mostly valid probably
                        return np.array(ordered_points), False
                    else:
                        ordered_points.append(best_next_pixel)
                    if ordered_points[-1] == ordered_points[-3] and ordered_points[-3] == ordered_points[-5]:
                        ordered_points = ordered_points[:-6]
                        return np.array(ordered_points), False
                    continue

        ordered_points.append(ordered_points[0])
        return np.array(ordered_points), True

    @staticmethod
    def circularTrace_old(trace_coordinates):

        """Reorders the coordinates of a trace from a circular DNA molecule
        (with no loops) using a polar coordinate system with reference to the
        center of mass

        I think every step of this can be vectorised for speed up

        This is vulnerable to bugs if the dna molecule folds in on itself slightly"""

        # calculate the centre of mass for the trace
        com_x = np.average(trace_coordinates[:, 0])
        com_y = np.average(trace_coordinates[:, 1])

        # convert to polar coordinates with respect to the centre of mass
        polar_coordinates = []
        for x1, y1 in trace_coordinates:

            x = x1 - com_x
            y = y1 - com_y

            r = math.hypot(x, y)
            theta = math.atan2(x, y)

            polar_coordinates.append([theta, r])

        sorted_polar_coordinates = sorted(polar_coordinates, key=lambda i: i[0])

        # Reconvert to x, y coordinates
        sorted_coordinates = []
        for theta, r in sorted_polar_coordinates:

            x = r * math.sin(theta)
            y = r * math.cos(theta)

            x2 = x + com_x
            y2 = y + com_y

            sorted_coordinates.append([x2, y2])

        return np.array(sorted_coordinates)

    def loopedCircularTrace():
        pass

    def loopedLinearTrace():
        pass


class genTracingFuncs:
    @staticmethod
    def getLocalPixelsBinary(binary_map, x, y):
        p2 = binary_map[x, y + 1]
        p3 = binary_map[x + 1, y + 1]
        p4 = binary_map[x + 1, y]
        p5 = binary_map[x + 1, y - 1]
        p6 = binary_map[x, y - 1]
        p7 = binary_map[x - 1, y - 1]
        p8 = binary_map[x - 1, y]
        p9 = binary_map[x - 1, y + 1]

        return p2, p3, p4, p5, p6, p7, p8, p9

    @staticmethod
    def countNeighbours(x, y, trace_coordinates):

        """Counts the number of neighbouring points for a given coordinate in
        a list of points"""

        number_of_neighbours = 0
        if [x, y + 1] in trace_coordinates:
            number_of_neighbours += 1
        if [x + 1, y + 1] in trace_coordinates:
            number_of_neighbours += 1
        if [x + 1, y] in trace_coordinates:
            number_of_neighbours += 1
        if [x + 1, y - 1] in trace_coordinates:
            number_of_neighbours += 1
        if [x, y - 1] in trace_coordinates:
            number_of_neighbours += 1
        if [x - 1, y - 1] in trace_coordinates:
            number_of_neighbours += 1
        if [x - 1, y] in trace_coordinates:
            number_of_neighbours += 1
        if [x - 1, y + 1] in trace_coordinates:
            number_of_neighbours += 1
        return number_of_neighbours

    @staticmethod
    def getNeighbours(x, y, trace_coordinates):

        """Returns an array containing the neighbouring points for a given
        coordinate in a list of points"""

        neighbour_array = []
        if [x, y + 1] in trace_coordinates:
            neighbour_array.append([x, y + 1])
        if [x + 1, y + 1] in trace_coordinates:
            neighbour_array.append([x + 1, y + 1])
        if [x + 1, y] in trace_coordinates:
            neighbour_array.append([x + 1, y])
        if [x + 1, y - 1] in trace_coordinates:
            neighbour_array.append([x + 1, y - 1])
        if [x, y - 1] in trace_coordinates:
            neighbour_array.append([x, y - 1])
        if [x - 1, y - 1] in trace_coordinates:
            neighbour_array.append([x - 1, y - 1])
        if [x - 1, y] in trace_coordinates:
            neighbour_array.append([x - 1, y])
        if [x - 1, y + 1] in trace_coordinates:
            neighbour_array.append([x - 1, y + 1])
        return neighbour_array

    @staticmethod
    def countandGetNeighbours(x, y, trace_coordinates):

        """Returns the number of neighbouring points for a coordinate and an
        array containing the those points"""

        neighbour_array = []
        number_of_neighbours = 0
        if [x, y + 1] in trace_coordinates:
            neighbour_array.append([x, y + 1])
            number_of_neighbours += 1
        if [x + 1, y + 1] in trace_coordinates:
            neighbour_array.append([x + 1, y + 1])
            number_of_neighbours += 1
        if [x + 1, y] in trace_coordinates:
            neighbour_array.append([x + 1, y])
            number_of_neighbours += 1
        if [x + 1, y - 1] in trace_coordinates:
            neighbour_array.append([x + 1, y - 1])
            number_of_neighbours += 1
        if [x, y - 1] in trace_coordinates:
            neighbour_array.append([x, y - 1])
            number_of_neighbours += 1
        if [x - 1, y - 1] in trace_coordinates:
            neighbour_array.append([x - 1, y - 1])
            number_of_neighbours += 1
        if [x - 1, y] in trace_coordinates:
            neighbour_array.append([x - 1, y])
            number_of_neighbours += 1
        if [x - 1, y + 1] in trace_coordinates:
            neighbour_array.append([x - 1, y + 1])
            number_of_neighbours += 1
        return number_of_neighbours, neighbour_array

    @staticmethod
    def returnPointsInArray(points_array, trace_coordinates):

        for x, y in points_array:
            if [x, y] in trace_coordinates:
                try:
                    points_in_trace_coordinates.append([x, y])
                except NameError:
                    points_in_trace_coordinates = [[x, y]]
        # for x, y in points_array:
        #    print([x,y])
        #    try:
        #        trace_coordinates.index([x,y])
        #        print(trace_coordinates.index([x,y]))
        #    except ValueError:
        #        continue
        #    else:
        #        try:
        #            points_in_trace_coordinates.append([x,y])
        #        except NameError:
        #            points_in_trace_coordinates = [[x,y]]
        try:
            return points_in_trace_coordinates
        except UnboundLocalError:
            return None

    @staticmethod
    def makeGrid(x, y, size):
        for x_n in range(-size, size + 1):
            x_2 = x + x_n
            for y_n in range(-size, size + 1):
                y_2 = y + y_n
                try:
                    grid.append([x_2, y_2])
                except NameError:
                    grid = [[x_2, y_2]]
        return grid

    @staticmethod
    def findBestNextPoint(x, y, ordered_points, candidate_points):

        ordered_points = np.array(ordered_points)
        candidate_points = np.array(candidate_points)

        ordered_points = ordered_points.tolist()
        candidate_points = candidate_points.tolist()

        for i in range(1, 8):
            # build array of coordinates from which to check
            coords_to_check = genTracingFuncs.makeGrid(x, y, i)
            # check for potential points in the larger search area
            points_in_array = genTracingFuncs.returnPointsInArray(coords_to_check, candidate_points)

            # Make a decision depending on how many points are found
            if not points_in_array:
                continue
            elif len(points_in_array) == 1:
                best_next_point = points_in_array[0]
                return best_next_point
            else:
                best_next_point = genTracingFuncs.checkVectorsCandidatePoints(x, y, ordered_points, points_in_array)
                return best_next_point
        return None

    @staticmethod
    def checkVectorsCandidatePoints(x, y, ordered_points, candidate_points):

        """Finds which neighbouring pixel incurs the smallest angular change
        with reference to a previous pixel in the ordered trace, and chooses that
        as the next point"""

        x_test = ordered_points[-1][0]
        y_test = ordered_points[-1][1]

        if len(ordered_points) > 4:
            x_ref = ordered_points[-3][0]
            y_ref = ordered_points[-3][1]

            x_ref_2 = ordered_points[-2][0]
            y_ref_2 = ordered_points[-2][1]
        elif len(ordered_points) > 3:
            x_ref = ordered_points[-2][0]
            y_ref = ordered_points[-2][1]

            x_ref_2 = ordered_points[0][0]
            y_ref_2 = ordered_points[0][1]
        else:
            x_ref = ordered_points[0][0]
            y_ref = ordered_points[0][1]

            x_ref_2 = ordered_points[0][0]
            y_ref_2 = ordered_points[0][1]

        dx = x_test - x_ref
        dy = y_test - y_ref

        ref_theta = math.atan2(dx, dy)

        x_y_theta = []

        for x_n, y_n in candidate_points:

            x = x_n - x_ref_2
            y = y_n - y_ref_2

            theta = math.atan2(x, y)

            x_y_theta.append([x_n, y_n, abs(theta - ref_theta)])

        ordered_x_y_theta = sorted(x_y_theta, key=lambda x: x[2])
        return [ordered_x_y_theta[0][0], ordered_x_y_theta[0][1]]



"""Custom Bruker Nanoscope colorscale."""
import logging
import matplotlib as mpl
from matplotlib.colors import LinearSegmentedColormap
import numpy as np

from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)


class Colormap:
    """Class for setting the COlormap"""

    def __init__(self, name: str = "nanoscope"):
        self.name = name
        self.cmap = None
        self.set(self.name)

    def __str__(self):
        return f"TopoStats Colormap: {self.name}"

    def set(self, name: str):
        """Set the ColorMap.

        Parameters
        ----------
        name: str
            Name of colormap
        """
        if name.lower() == "nanoscope":
            self.cmap = self.nanoscope()
        else:
            # Get one of the matplotlib colormaps
            self.cmap = mpl.colormaps[name]
        LOGGER.debug(f"[theme] Colormap set to : {name}")

    def get_cmap(self):
        """Return the matplotlib.cm colormap object"""
        return self.cmap

    @staticmethod
    def nanoscope():
        """
        Returns a matplotlib compatible colormap that replicates the Bruker Nanoscope colorscale
        The colormap is implemented in Gwyddion's GwyGradient via 'Nanoscope.txt'

        Original Nanoscope.txt

        Gwyddion resource GwyGradient
        0 0 0 0 1
        0.124464 0 0 0 1
        0.236052 0.0670103 0 0 1
        0.371245 0.253338 0 0 1
        0.472103 0.392344 0.0721649 0 1
        0.611588 0.584587 0.334114 0 1
        0.708155 0.717678 0.515464 0.252575 1
        0.714052 0.725806 0.527471 0.268 1
        0.890558 0.969072 0.886843 0.76343 1
        0.933476 0.987464 0.974227 0.883897 1
        0.944709 0.992278 0.980523 0.915426 1
        0.965682 0.995207 0.992278 0.974293 1
        0.971401 0.996006 0.993565 0.990347 1
        1 1 1 1 1
        """

        N = 9  # Number of values
        vals = np.ones((N, 4))  # Initialise the array to be full of 1.0
        vals[0] = [0.0, 0.0, 0.0, 1]
        vals[1] = [0.0670103, 0.0, 0.0, 1.0]
        vals[2] = [0.07, 0.01, 0.01, 1.0]
        vals[3] = [0.14, 0.02, 0.0, 1.0]
        vals[4] = [0.38, 0.1, 0.0, 1.0]
        vals[5] = [0.51, 0.30, 0.11, 1.0]
        vals[6] = [0.78, 0.56, 0.42, 1.0]
        vals[7] = [0.88, 0.87, 0.69, 1.0]
        vals[8] = [1.0, 1.0, 1.0, 1.0]

        colormap = LinearSegmentedColormap.from_list("nanoscope", vals, N=256)
        return colormap



"""Contains filter functions that take a 2D array representing an image as an input, as well as necessary parameters,
and return a 2D array of the same size representing the filtered image."""
import logging

from skimage.filters import gaussian
import numpy as np

from topostats.logs.logs import LOGGER_NAME
from topostats.utils import get_thresholds, get_mask

LOGGER = logging.getLogger(LOGGER_NAME)

# pylint: disable=fixme
# pylint: disable=broad-except


class Filters:
    """Class for filtering scans."""

    def __init__(
        self,
        image: np.ndarray,
        filename: str,
        pixel_to_nm_scaling: float,
        threshold_method: str = "otsu",
        otsu_threshold_multiplier: float = 1.7,
        threshold_std_dev: dict = None,
        threshold_absolute: dict = None,
        gaussian_size: float = None,
        gaussian_mode: str = "nearest",
        quiet: bool = False,
    ):
        """Initialise the class.

        Parameters
        ----------
        image: np.ndarray
            The raw image from the AFM.
        filename: str
            The filename (used for logging outputs only).
        pixel_to_nm_scaling: float
            Value for converting pixels to nanometers.
        threshold_method: str
            Method for thresholding, default 'otsu', valid options 'otsu', 'std_dev' and 'absolute'.
        otsu_threshold_multiplier: float
            Value for scaling the derived Otsu threshold (optional).
        threshold_std_dev: dict
            If using the 'std_dev' threshold method. Dictionary that contains upper and lower
            threshold values for the number of standard deviations from the mean to threshold.
        threshold_absolute: dict
            If using the 'absolute' threshold method. Dictionary that contains upper and lower
            absolute threshold values for flattening.
        quiet: bool
            Whether to silence output.
        """
        self.filename = filename
        self.pixel_to_nm_scaling = pixel_to_nm_scaling
        self.gaussian_size = gaussian_size
        self.gaussian_mode = gaussian_mode
        self.threshold_method = threshold_method
        self.otsu_threshold_multiplier = otsu_threshold_multiplier
        self.threshold_std_dev = threshold_std_dev
        self.threshold_absolute = threshold_absolute
        self.images = {
            "pixels": image,
            "initial_median_flatten": None,
            "initial_tilt_removal": None,
            "initial_quadratic_removal": None,
            "masked_median_flatten": None,
            "masked_tilt_removal": None,
            "masked_quadratic_removal": None,
            "mask": None,
            "zero_average_background": None,
            "gaussian_filtered": None,
        }
        self.thresholds = None
        self.medians = {"rows": None, "cols": None}
        self.results = {
            "diff": None,
            "median_row_height": None,
            "x_gradient": None,
            "y_gradient": None,
            "threshold": None,
        }

        if quiet:
            LOGGER.setLevel("ERROR")

    def median_flatten(self, image: np.ndarray, mask: np.ndarray = None, img_name: str = None) -> np.ndarray:
        """
        Uses the method of median differences to flatten the rows of an image, aligning the rows and centering the
        median around zero. When used with a mask, this has the effect of centering the background data on zero.
        Note this function does not handle scars.

        Parameters
        ----------
        image: np.ndarray
            2-D image of the data to align the rows of.
        mask: np.ndarray
            Boolean array of points to mask out (ignore).
        img_name: str
            Name of the image (to be able to print information in the console).
        Returns
        -------
        np.ndarray
            Returns a copy of the input image with rows aligned
        """
        image = image.copy()
        if mask is not None:
            read_matrix = np.ma.masked_array(image, mask=mask, fill_value=np.nan).filled()
            LOGGER.info(f"[{img_name}] : Median flattening with mask")
        else:
            read_matrix = image
            LOGGER.info(f"[{img_name}] : Median flattening without mask")

        for row in range(image.shape[0]):
            # Get the median of the row
            m = np.nanmedian(read_matrix[row, :])
            if not np.isnan(m):
                image[row, :] -= m
            else:
                LOGGER.warning(
                    """f[{self.filename}] Large grain detected image can not be
processed, please refer to <url to page where we document common problems> for more information."""
                )

        return image

    def remove_tilt(self, image: np.ndarray, mask: np.ndarray = None, img_name: str = None):
        """
        Removes planar tilt from an image (linear in 2D space). It uses a linear fit of the medians
        of the rows and columns to determine the linear slants in x and y directions and then subtracts the fit from the columns.

        Parameters
        ----------
        image: np.ndarray
            2-D image of the data to remove the planar tilt from.
        mask: np.ndarray
            Boolean array of points to mask out (ignore).
        img_name: str
            Name of the image (to be able to print information in the console).
        Returns
        -------
        np.ndarray
            Returns a copy of the input image with the planar tilt removed
        """
        image = image.copy()
        if mask is not None:
            read_matrix = np.ma.masked_array(image, mask=mask, fill_value=np.nan).filled()
            LOGGER.info(f"[{img_name}] : Plane tilt removal with mask")
        else:
            read_matrix = image
            LOGGER.info(f"[{img_name}] : Plane tilt removal without mask")

        # Line of best fit
        # Calculate medians
        medians_x = [np.nanmedian(read_matrix[:, i]) for i in range(read_matrix.shape[1])]
        medians_y = [np.nanmedian(read_matrix[j, :]) for j in range(read_matrix.shape[0])]

        # Fit linear x
        px = np.polyfit(range(0, len(medians_x)), medians_x, 1)
        LOGGER.info(f"x-polyfit 1st order: {px}")
        py = np.polyfit(range(0, len(medians_y)), medians_y, 1)
        LOGGER.info(f"y-polyfit 1st order: {py}")

        if px[0] != 0:
            if not np.isnan(px[0]):
                LOGGER.info("removing x plane tilt")
                for row in range(0, image.shape[0]):
                    for col in range(0, image.shape[1]):
                        image[row, col] -= px[0] * (col)
            else:
                LOGGER.info("x gradient is nan, skipping plane tilt x removal")
        else:
            LOGGER.info("x gradient is zero, skipping plane tilt x removal")

        if py[0] != 0:
            if not np.isnan(py[0]):
                LOGGER.info("removing y plane tilt")
                for row in range(0, image.shape[0]):
                    for col in range(0, image.shape[1]):
                        image[row, col] -= py[0] * (row)
            else:
                LOGGER.info("y gradient is nan, skipping plane tilt y removal")
        else:
            LOGGER.info("y gradient is zero, skipping plane tilt y removal")

        return image

    def remove_quadratic(self, image: np.ndarray, mask: np.ndarray = None, img_name: str = None):
        """
        Removes the quadratic bowing that can be seen in some large-scale AFM images. It uses a simple quadratic fit
        on the medians of the columns of the image and then subtracts the calculated quadratic from the columns.

        Parameters
        ----------
        image: np.ndarray
            2-D image of the data to remove the quadratic from.
        mask: np.ndarray
            Boolean array of points to mask out (ignore).
        img_name: str
            Name of the image (to be able to print information in the console).
        Returns
        -------
        np.ndarray
            Returns a copy of the input image with the quadratic bowing removed
        """
        image = image.copy()
        if mask is not None:
            read_matrix = np.ma.masked_array(image, mask=mask, fill_value=np.nan).filled()
            LOGGER.info(f"[{img_name}] : Remove quadratic bow with mask")
        else:
            read_matrix = image
            LOGGER.info(f"[{img_name}] : Remove quadratic bow without mask")

        # Calculate medians
        medians_x = [np.nanmedian(read_matrix[:, i]) for i in range(read_matrix.shape[1])]

        # Fit quadratic x
        px = np.polyfit(range(0, len(medians_x)), medians_x, 2)
        LOGGER.info(f"x polyfit 2nd order: {px}")

        # Handle divide by zero
        if px[0] != 0:
            if not np.isnan(px[0]):
                # Remove quadratic in x
                cx = -px[1] / (2 * px[0])
                for row in range(0, image.shape[0]):
                    for col in range(0, image.shape[1]):
                        image[row, col] -= px[0] * (col - cx) ** 2
            else:
                LOGGER.info("quadratic polyfit returns nan, skipping quadratic removal")
        else:
            LOGGER.info("quadratic polyfit returns zero, skipping quadratic removal")

        return image

    @staticmethod
    def calc_diff(array: np.ndarray) -> np.ndarray:
        """Calculate the difference of an array."""
        return array[-1] - array[0]

    def calc_gradient(self, array: np.ndarray, shape: int) -> np.ndarray:
        """Calculate the gradient of an array."""
        return self.calc_diff(array) / shape

    def average_background(self, image: np.ndarray, mask: np.ndarray = None) -> np.ndarray:
        """Zero the background by subtracting the non-masked mean from all pixels.

        Parameters
        ----------
        image: np.array
            Numpy array representing image.
        mask: np.array
            Mask of the array, should have the same dimensions as image.

        Returns
        -------
        np.ndarray
            Numpy array of image zero averaged.
        """
        if mask is None:
            mask = np.zeros_like(image)
        mean = np.mean(image[mask == 0])
        LOGGER.info(f"[{self.filename}] : Zero averaging background : {mean} nm")
        return image - mean

    def gaussian_filter(self, image: np.ndarray, **kwargs) -> np.array:
        """Apply Gaussian filter to an image.

        Parameters
        ----------
        image: np.array
            Numpy array representing image.

        Returns
        -------
        np.array
            Numpy array of gaussian blurred image.

        """
        LOGGER.info(
            f"[{self.filename}] : Applying Gaussian filter (mode : {self.gaussian_mode};"
            f" Gaussian blur (px) : {self.gaussian_size})."
        )
        return gaussian(
            image,
            sigma=(self.gaussian_size),
            mode=self.gaussian_mode,
            **kwargs,
        )

    def filter_image(self) -> None:
        """Process a single image, filtering, finding grains and calculating their statistics.

        Example
        -------
        from topostats.io import LoadScan
        from topostats.topotracing import Filter, process_scan

        load_scan = LoadScan("minicircle.spm", channel="Height")
        load_scan.get_data()
        filter = Filter(image=load_scan.image,
        ...             pixel_to_nm_scaling=load_scan.pixel_to_nm_scaling,
        ...             filename=load_scan.filename,
        ...             threshold_method='otsu')
        filter.filter_image()

        """
        self.images["initial_median_flatten"] = self.median_flatten(
            self.images["pixels"], mask=None, img_name=self.filename
        )
        self.images["initial_tilt_removal"] = self.remove_tilt(
            self.images["initial_median_flatten"], mask=None, img_name=self.filename
        )
        self.images["initial_quadratic_removal"] = self.remove_quadratic(
            self.images["initial_tilt_removal"], mask=None, img_name=self.filename
        )

        # Get the thresholds
        try:
            self.thresholds = get_thresholds(
                image=self.images["initial_quadratic_removal"],
                threshold_method=self.threshold_method,
                otsu_threshold_multiplier=self.otsu_threshold_multiplier,
                threshold_std_dev=self.threshold_std_dev,
                absolute=self.threshold_absolute,
            )
        except TypeError as type_error:
            raise type_error
        self.images["mask"] = get_mask(
            image=self.images["initial_quadratic_removal"], thresholds=self.thresholds, img_name=self.filename
        )
        self.images["masked_median_flatten"] = self.median_flatten(
            self.images["initial_tilt_removal"], self.images["mask"], img_name=self.filename
        )
        self.images["masked_tilt_removal"] = self.remove_tilt(
            self.images["masked_median_flatten"], self.images["mask"], img_name=self.filename
        )
        self.images["masked_quadratic_removal"] = self.remove_quadratic(
            self.images["masked_tilt_removal"], self.images["mask"], img_name=self.filename
        )
        self.images["zero_average_background"] = self.average_background(
            self.images["masked_quadratic_removal"], self.images["mask"]
        )
        self.images["gaussian_filtered"] = self.gaussian_filter(self.images["zero_average_background"])



"""Functions for reading and writing data."""
import logging
from datetime import datetime
from pathlib import Path
from typing import Union, Dict
import numpy as np

from pySPM.Bruker import Bruker
from igor import binarywave  # this is already imported with afmformats but is being specialised for our use
from afmformats import load_data
from afmformats.mod_creep_compliance import AFMCreepCompliance
from ruamel.yaml import YAML, YAMLError
from ruamel.yaml.main import round_trip_load as yaml_load, round_trip_dump as yaml_dump
from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)

# pylint: disable=broad-except


def read_yaml(filename: Union[str, Path]) -> Dict:
    """Read a YAML file.

    Parameters
    ----------
    filename: Union[str, Path]
        YAML file to read.

    Returns
    -------
    Dict
        Dictionary of the file."""

    with Path(filename).open(encoding="utf-8") as f:
        try:
            yaml_file = YAML(typ="safe")
            return yaml_file.load(f)
        except YAMLError as exception:
            LOGGER.error(exception)
            return {}


def write_yaml(config: dict, output_dir: Union[str, Path]) -> None:
    """Write a configuration (stored as a dictionary) to a YAML file.

    Parameters
    ----------
    config: dict
        Configuration dictionary.
    output_dir: Union[str, Path]
        Path to save the dictionary to as a YAML file (it will be called 'config.yaml').
    """
    # Save the configuration to output directory
    output_config = Path(output_dir) / "config.yaml"
    # Revert PosixPath items to string
    config["base_dir"] = str(config["base_dir"])
    config["output_dir"] = str(config["output_dir"])
    config_yaml = yaml_load(yaml_dump(config))
    config_yaml.yaml_set_start_comment(
        f"Configuration from TopoStats run completed : {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
    )
    with output_config.open("w") as f:
        try:
            f.write(yaml_dump(config_yaml))
        except YAMLError as exception:
            LOGGER.error(exception)


class LoadScan:
    """Load the image and image parameters from a file path."""

    def __init__(
        self,
        img_path: Union[str, Path],
        channel: str,
    ):

        """Initialise the class.

        Parameters
        ----------
        img_path: Union[str, Path]
            Path to a valid AFM scan to load.
        channel: str
            Image channel to extract from the scan.
        """
        self.img_path = Path(img_path)
        self.channel = channel
        self.channel_data = None
        self.file_path = self.img_path
        self.filename = self.img_path.stem
        self.suffix = self.img_path.suffix
        self.image = None
        self.pixel_to_nm_scaling = None

    def load_spm(self) -> tuple:
        """Extract image and pixel to nm scaling from the Bruker .spm file."""
        LOGGER.info(f"Loading image from : {self.img_path}")
        try:
            scan = Bruker(self.img_path)
            LOGGER.info(f"[{self.filename}] : Loaded image from : {self.img_path}")
            self.channel_data = scan.get_channel(self.channel)
            LOGGER.info(f"[{self.filename}] : Extracted channel {self.channel}")
            image = np.flipud(np.array(self.channel_data.pixels))
        except FileNotFoundError:
            LOGGER.info(f"[{self.filename}] File not found : {self.img_path}")
        except Exception:
            # trying to return the error with options of possible channel values
            labels = []
            for channel in [layer[b"@2:Image Data"][0] for layer in scan.layers]:
                channel_name = channel.decode("latin1").split(" ")[1][1:-1]
                # channel_description = channel.decode('latin1').split('"')[1] # incase the blank field raises quesions?
                labels.append(channel_name)
            LOGGER.error(f"[{self.filename}] : {self.channel} not in {self.suffix} channel list: {labels}")
            raise

        return (image, self._spm_pixel_to_nm_scaling(self.channel_data))

    def _spm_pixel_to_nm_scaling(self, channel_data) -> float:
        """Extract pixel to nm scaling from the SPM image metadata.

        Parameters
        ----------
        channel_data:
            Channel data

        Returns
        -------
        float
            Pixel to nm scaling factor.
        """
        unit_dict = {
            "nm": 1,
            "um": 1e3,
        }
        px_to_real = channel_data.pxs()
        # Has potential for non-square pixels but not yet implimented
        pixel_to_nm_scaling = (
            px_to_real[0][0] * unit_dict[px_to_real[0][1]],
            px_to_real[1][0] * unit_dict[px_to_real[1][1]],
        )[0]
        if px_to_real[0][0] == 0 and px_to_real[1][0] == 0:
            pixel_to_nm_scaling = 1
            LOGGER.warning(f"[{self.filename}] : Pixel size not found in metadata, defaulting to 1nm")
        LOGGER.info(f"[{self.filename}] : Pixel to nm scaling : {pixel_to_nm_scaling}")
        return pixel_to_nm_scaling

    # def load_jpk(self) -> None:
    #     """Load and extract image from .jpk files."""
    #     jpk = self._load_jpk()
    #     data = self._extract_jpk(jpk)
    #     return (jpk, None)

    def load_ibw(self) -> tuple:
        """Loads image from Asylum Research (Igor) .ibw files"""

        LOGGER.info(f"Loading image from : {self.img_path}")
        try:
            scan = binarywave.load(self.img_path)
            LOGGER.info(f"[{self.filename}] : Loaded image from : {self.img_path}")

            labels = []
            for label_list in scan["wave"]["labels"]:
                for label in label_list:
                    if label:
                        labels.append(label.decode())
            channel_idx = labels.index(self.channel)
            image = scan["wave"]["wData"][:, :, channel_idx].T * 1e9  # Looks to be in m
            image = np.flipud(image)
            LOGGER.info(f"[{self.filename}] : Extracted channel {self.channel}")
        except FileNotFoundError:
            LOGGER.info(f"[{self.filename}] File not found : {self.img_path}")
        except ValueError:
            LOGGER.error(f"[{self.filename}] : {self.channel} not in {self.suffix} channel list: {labels}")
            raise
        except Exception as exception:
            LOGGER.error(f"[{self.filename}] : {exception}")

        return (image, self._ibw_pixel_to_nm_scaling(scan))

    def _ibw_pixel_to_nm_scaling(self, scan) -> float:
        """Extract pixel to nm scaling from the IBW image metadata."""
        # Get metadata
        notes = {}
        for line in str(scan["wave"]["note"]).split("\\r"):
            if line.count(":"):
                key, val = line.split(":", 1)
                notes[key] = val.strip()
        # Has potential for non-square pixels but not yet implimented
        pixel_to_nm_scaling = (
            float(notes["SlowScanSize"]) / scan["wave"]["wData"].shape[0] * 1e9,  # as in m
            float(notes["FastScanSize"]) / scan["wave"]["wData"].shape[1] * 1e9,  # as in m
        )[0]
        LOGGER.info(f"[{self.filename}] : Pixel to nm scaling : {pixel_to_nm_scaling}")
        return pixel_to_nm_scaling

    def load_jpk(self) -> None:
        try:
            jpk = load_data(self.img_path)
        except FileNotFoundError:
            LOGGER.info(f"[{self.filename}] File not found : {self.img_path}")
        except Exception as exception:
            LOGGER.error(f"[{self.filename}] : {exception}")
        return jpk

    @staticmethod
    def _extract_jpk(jpk: AFMCreepCompliance) -> np.ndarray:
        """Extract data from jpk object."""

    def get_data(self) -> None:
        """Method to extract image and pixel to nm scaling."""
        LOGGER.info(f"Extracting image from {self.suffix}")
        if self.suffix == ".spm":
            self.image, self.pixel_to_nm_scaling = self.load_spm()
        if self.suffix == ".jpk":
            self.image, self.pixel_to_nm_scaling = self.load_jpk()
        if self.suffix == ".ibw":
            self.image, self.pixel_to_nm_scaling = self.load_ibw()



"""Functions for calculating thresholds."""
# pylint: disable=no-name-in-module
import logging
from typing import Callable
import numpy as np
from skimage.filters import threshold_mean, threshold_minimum, threshold_otsu, threshold_yen, threshold_triangle

from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)

# pylint: disable=no-else-return


def threshold(image: np.ndarray, method: str = None, otsu_threshold_multiplier: float = None, **kwargs: dict) -> float:
    """Factory method for thresholding.

    Parameters
    ----------
    method : str
        Method to use for thresholding, currently supported methods are otsu (default), mean and minimum.
    **kwargs : dict
        Additional keyword arguments to pass to skimage methods.

    Returns
    -------
    float
        Threshold of image using specified method.
    """
    thresholder = _get_threshold(method)
    return thresholder(image, otsu_threshold_multiplier=otsu_threshold_multiplier, **kwargs)


def _get_threshold(method: str = "otsu") -> Callable:
    """Creator component which determines which threshold method to use.

    Parameters
    ----------
    method : str
        Threshold method to use, currently supports otsu (default), std_dev_lower, std_dev_upper, minimum, mean and yen.

    Returns
    -------
    function
        Returns function appropriate for the required threshold method.

    Raises
    ------
    ValueError
        Unsupported methods result in ValueError.
    """
    if method == "otsu":
        return _threshold_otsu
    if method == "mean":
        return _threshold_mean
    if method == "minimum":
        return _threshold_minimum
    if method == "yen":
        return _threshold_yen
    if method == "triangle":
        return _threshold_triangle
    raise ValueError(method)


def _threshold_otsu(image: np.ndarray, otsu_threshold_multiplier: float = None, **kwargs) -> float:
    return threshold_otsu(image, **kwargs) * otsu_threshold_multiplier


def _threshold_mean(image: np.ndarray, otsu_threshold_multiplier: float = None, **kwargs) -> float:
    return threshold_mean(image, **kwargs)


def _threshold_minimum(image: np.ndarray, otsu_threshold_multiplier: float = None, **kwargs) -> float:
    return threshold_minimum(image, **kwargs)


def _threshold_yen(image: np.ndarray, otsu_threshold_multiplier: float = None, **kwargs) -> float:
    return threshold_yen(image, **kwargs)


def _threshold_triangle(image: np.ndarray, otsu_threshold_multiplier: float = None, **kwargs) -> float:
    return threshold_triangle(image, **kwargs)



"""Find grains in an image."""
# pylint: disable=no-name-in-module
from collections import defaultdict
import logging
from typing import List, Dict
import numpy as np

from skimage.segmentation import clear_border
from skimage.morphology import remove_small_objects, label
from skimage.measure import regionprops
from skimage.color import label2rgb

from topostats.logs.logs import LOGGER_NAME
from topostats.thresholds import threshold
from topostats.utils import _get_mask, get_thresholds

LOGGER = logging.getLogger(LOGGER_NAME)

# pylint: disable=fixme
# pylint: disable=line-too-long
# pylint: disable=too-many-instance-attributes
# pylint: disable=bare-except


class Grains:
    """Find grains in an image."""

    def __init__(
        self,
        image: np.ndarray,
        filename: str,
        pixel_to_nm_scaling: float,
        threshold_method: str = None,
        otsu_threshold_multiplier: float = None,
        threshold_std_dev: dict = None,
        threshold_absolute: dict = None,
        absolute_area_threshold: dict = {
            "upper": [None, None],
            "lower": [None, None],
        },
        direction: str = None,
        absolute_smallest_grain_size: float = None,
    ):
        """Initialise the class.

        Parameters
        ----------
        image: np.ndarray
            2D Numpy array of image
        filename: str
            File being processed
        pixel_to_nm_scaling: float
            Sacling of pixels to nanometre.
        threshold_multiplier : Union[int, float]
            Factor by which lower threshold is to be scaled prior to masking.
        threshold_method: str
            Method for determining threshold to mask values, default is 'otsu'.
        threshold_std_dev: dict
            Dictionary of 'lower' and 'upper' factors by which standard deviation is multiplied to derive the threshold if threshold_method is 'std_dev'.
        threshold_absolute: dict
            Dictionary of absolute 'lower' and 'upper' thresholds for grain finding.
        absolute_area_threshold: dict
            Dictionary of upper and lower grain's area thresholds
        direction: str
            Direction for which grains are to be detected, valid values are upper, lower and both.
        """
        self.image = image
        self.filename = filename
        self.pixel_to_nm_scaling = pixel_to_nm_scaling
        self.threshold_method = threshold_method
        self.otsu_threshold_multiplier = otsu_threshold_multiplier
        self.threshold_std_dev = threshold_std_dev
        self.threshold_absolute = threshold_absolute
        self.absolute_area_threshold = absolute_area_threshold
        # Only detect grains for the desired direction
        self.direction = [direction] if direction != "both" else ["upper", "lower"]
        self.absolute_smallest_grain_size = absolute_smallest_grain_size
        self.thresholds = None
        self.images = {
            "mask_grains": None,
            "tidied_border": None,
            "tiny_objects_removed": None,
            "objects_removed": None,
            # "labelled_regions": None,
            # "coloured_regions": None,
        }
        self.directions = defaultdict()
        self.minimum_grain_size = None
        self.region_properties = defaultdict()
        self.bounding_boxes = defaultdict()
        self.grainstats = None

    def tidy_border(self, image: np.array, **kwargs) -> np.array:
        """Remove grains touching the border.

        Parameters
        ----------
        image: np.array
            Numpy array representing image.

        Returns
        -------
        np.array
            Numpy array of image with borders tidied.
        """
        LOGGER.info(f"[{self.filename}] : Tidying borders")
        return clear_border(image, **kwargs)

    def label_regions(self, image: np.array) -> np.array:
        """Label regions.

        This method is used twice, once prior to removal of small regions, and again afterwards, hence requiring an
        argument of what image to label.

        Parameters
        ----------
        image: np.array
            Numpy array representing image.

        Returns
        -------
        np.array
            Numpy array of image with objects coloured.
        """
        LOGGER.info(f"[{self.filename}] : Labelling Regions")
        return label(image, background=0)

    def calc_minimum_grain_size(self, image: np.ndarray) -> float:
        """Calculate the minimum grain size.

        Very small objects are first removed via thresholding before calculating the lower extreme.
        """
        region_properties = self.get_region_properties(image)
        grain_areas = np.array([grain.area for grain in region_properties])
        if len(grain_areas > 0):
            # Exclude small objects less than a given threshold first
            grain_areas = grain_areas[
                grain_areas >= threshold(grain_areas, method="otsu", otsu_threshold_multiplier=1.0)
            ]
            self.minimum_grain_size = np.median(grain_areas) - (
                1.5 * (np.quantile(grain_areas, 0.75) - np.quantile(grain_areas, 0.25))
            )
        else:
            self.minimum_grain_size = -1

    def remove_noise(self, image: np.ndarray) -> np.ndarray:
        """Removes noise which are objects smaller than the 'absolute_smallest_grain_size'.

        This ensures that the smallest objects ~1px are removed regardless of the size distribution of the grains.

        Parameters
        ----------
        image: np.ndarray
            2D Numpy image to be cleaned.

        Returns
        -------
        np.ndarray
            2D Numpy array of image with objects < absolute_smallest_grain_size removed.
        """
        LOGGER.info(f"[{self.filename}] : Removing noise (< {self.absolute_smallest_grain_size})")
        return remove_small_objects(image, min_size=self.absolute_smallest_grain_size)

    def remove_small_objects(self, image: np.array, **kwargs):
        """Remove small objects."""
        # If self.minimum_grain_size is -1, then this means that
        # there were no grains to calculate the minimum grian size from.
        if self.minimum_grain_size != -1:
            small_objects_removed = remove_small_objects(
                image,
                min_size=(self.minimum_grain_size * self.pixel_to_nm_scaling),
                **kwargs,
            )
            LOGGER.info(
                f"[{self.filename}] : Removed small objects (< {self.minimum_grain_size * self.pixel_to_nm_scaling})"
            )
            return small_objects_removed > 0.0
        return image

    def area_thresholding(self, image: np.ndarray, area_thresholds: list):
        """Removes objects larger and smaller than the specified thresholds.

        Parameters
        ----------
        image: np.ndarray
            Image array where the background == 0 and grains are labelled as integers > 0.
        area_thresholds: list
            List of area thresholds, first should be the lower threshold, second upper threshold.

        Returns
        -------
        np.ndarray
            Image where grains outside the thresholds have been removed, as a re-numbered labeled image.

        """
        image_cp = image.copy()
        lower, upper = area_thresholds
        # if one value is None adjust for comparison
        if upper is None:
            upper = image.size * self.pixel_to_nm_scaling**2
        if lower is None:
            lower = 0
        uniq = np.delete(np.unique(image), 0)
        grain_count = 0
        for grain_no in uniq:
            grain_area = np.sum(image_cp == grain_no) * (self.pixel_to_nm_scaling**2)
            if grain_area > upper or grain_area < lower:
                image_cp[image_cp == grain_no] = 0
            else:
                grain_count += 1
                image_cp[image_cp == grain_no] = grain_count
        return image_cp

    def colour_regions(self, image: np.array, **kwargs) -> np.array:
        """Colour the regions.

        Parameters
        ----------
        image: np.array
            Numpy array representing image.

        Returns
        -------
        np.array
            Numpy array of image with objects coloured.
        """

        coloured_regions = label2rgb(image, **kwargs)
        LOGGER.info(f"[{self.filename}] : Coloured regions")
        return coloured_regions

    @staticmethod
    def get_region_properties(image: np.array, **kwargs) -> List:
        """Extract the properties of each region.

        Parameters
        ----------
        image: np.array
            Numpy array representing image

        Returns
        -------
        List
            List of region property objects.
        """
        return regionprops(image, **kwargs)

    def get_bounding_boxes(self, direction) -> Dict:
        """Derive a list of bounding boxes for each region from the derived region_properties

        Parameters
        ----------
        direction: str
            Direction of threshold for which bounding boxes are being calculated.

        Returns
        -------
        dict
            Dictionary of bounding boxes indexed by region area.
        """
        return {region.area: region.area_bbox for region in self.region_properties[direction]}

    def find_grains(self):
        """Find grains."""
        LOGGER.info(f"[{self.filename}] : Thresholding method (grains) : {self.threshold_method}")
        self.thresholds = get_thresholds(
            image=self.image,
            threshold_method=self.threshold_method,
            otsu_threshold_multiplier=self.otsu_threshold_multiplier,
            threshold_std_dev=self.threshold_std_dev,
            absolute=self.threshold_absolute,
        )
        try:
            region_props_count = 0
            for direction in self.direction:
                LOGGER.info(f"[{self.filename}] : Processing {direction} threshold ({self.thresholds[direction]})")
                self.directions[direction] = defaultdict()
                self.directions[direction]["mask_grains"] = _get_mask(
                    self.image,
                    threshold=self.thresholds[direction],
                    threshold_direction=direction,
                    img_name=self.filename,
                )
                self.directions[direction]["labelled_regions_01"] = self.label_regions(
                    self.directions[direction]["mask_grains"]
                )
                self.directions[direction]["tidied_border"] = self.tidy_border(
                    self.directions[direction]["labelled_regions_01"]
                )
                self.directions[direction]["removed_noise"] = self.area_thresholding(
                    self.directions[direction]["tidied_border"],
                    [self.absolute_smallest_grain_size * self.pixel_to_nm_scaling, None],
                )
                # if no area thresholds specified, use otsu
                if self.absolute_area_threshold[direction].count(None) == 2:
                    self.calc_minimum_grain_size(self.directions[direction]["removed_noise"])
                    self.directions[direction]["removed_small_objects"] = self.remove_small_objects(
                        self.directions[direction]["removed_noise"]
                    )
                else:
                    self.directions[direction]["removed_small_objects"] = self.area_thresholding(
                        self.directions[direction]["removed_noise"],
                        self.absolute_area_threshold[direction],
                    )

                self.directions[direction]["labelled_regions_02"] = self.label_regions(
                    self.directions[direction]["removed_small_objects"]
                )
                self.region_properties[direction] = self.get_region_properties(
                    self.directions[direction]["labelled_regions_02"]
                )
                LOGGER.info(f"[{self.filename}] : Region properties calculated ({direction})")
                self.directions[direction]["coloured_regions"] = self.colour_regions(
                    self.directions[direction]["labelled_regions_02"]
                )
                self.bounding_boxes[direction] = self.get_bounding_boxes(direction=direction)
                LOGGER.info(f"[{self.filename}] : Extracted bounding boxes ({direction})")
                region_props_count += len(self.region_properties[direction])
            if region_props_count == 0:
                self.region_properties = None
        # FIXME : Identify what exception is raised with images without grains and replace broad except
        except:  # noqa: E722
            LOGGER.info(f"[{self.filename}] : No grains found.")
